/**
 ** David Bonnafous - Anthony Mallet - Thu Jul  5 2001
 ** Copyright (c) 2001 LAAS/CNRS 
 **
 ** $Log: arcs.c,v $
 ** Revision 1.2  2007/11/22 10:19:30  adegroot
 ** Compute a->depth for each arc.
 **
 ** While here, reindent the file
 **
 ** Revision 1.1.1.1  2006/10/17 18:09:42  sjoyeux
 ** Packaged the P3D module from Lama's calife.
 ** See CHANGES for a list of modifications
 **
 **
 ** Revision 1.5  2003/09/12 15:21:02  mallet
 ** correctly handle arcs that cross goalTol limit
 **
 ** Revision 1.4  2001/09/13 10:49:01  mallet
 ** le nombre d'arcs pris en compte est divise par 2 a chaque niveau de recursion. C'est suffisant et ca accelere beaucoup la recherche
 **
 ** Revision 1.3  2001/07/14 11:39:09  mallet
 ** premiere version ok
 **
 ** Revision 1.2  2001/07/12 17:38:45  mallet
 ** sync
 **
 ** Revision 1.1  2001/07/11 18:05:40  mallet
 ** re-ecriture de p3d
 **
 **/
#ident "$Id: arcs.c,v 1.2 2007/11/22 10:19:30 adegroot Exp $"

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>

#include "libp3d.h"
#include "p3dInt.h"

static P3DArc *
_p3dArcsCreate(int nArcs, int nNodes , double distance, double x, double y,
		double theta, double rMin, int depth, int local_depth);

/*
 * p3dArcsCreate --------------------------------------------------------
 *
 */

static P3DArc *
_p3dArcsCreate(int nArcs, int nNodes, double distance, double x, double y,
		double theta, double rMin, int depth, int local_depth)
		
{
	int i,j, nR;
	P3DArc *arcs, *a;
	double ds;

	nR = nArcs/2;
	arcs = malloc((2*nR+1)*sizeof(*arcs));
	if (arcs == NULL) { 
		errno = ENOMEM; 
		return NULL; 
	}

	a = arcs;
	ds = distance / nNodes; 
	for(i=-nR, a=arcs; i<=nR; i++, a++) {

		/* arc curvature */
		a->r = i/(rMin*nR);

		/* compute number of nodes and limit delta theta */
		a->n = nNodes;

		/* alloc nodes */
		a->nodes = malloc(a->n*sizeof(*(a->nodes)));
		if (a->nodes == NULL) {
			free(arcs);
			errno = ENOMEM;
			return NULL;
		}

		/* initialize */
		a->ds = ds;
		for(j=0;j<a->n;j++) {
			p3dGetArcPos(a->r, j*ds + ds, x, y, theta,
					&a->nodes[j].x, &a->nodes[j].y, &a->nodes[j].theta);

			a->nodes[j].conf.q = -1.0;
			a->nodes[j].prev = (j == 0) ? NULL : &a->nodes[j-1].conf;

			a->nodes[j].cost = P3D_UNDEF_COST;
			a->nodes[j].remain = P3D_UNDEF_COST;

			a->nodes[j].number = j;
			a->nodes[j].parent = a;
		}

		a->current = 0;
		a->is_elected = 0;
		a->first_unknow = 0;
		a->cost = P3D_UNDEF_COST;

		a->number = i;
		a->parent = NULL;
		a->depth = local_depth;

		if (depth <= 1) {
			a->nArcs = 0;
			a->arcs = NULL;
		} else {
			/* half number of arcs */
			a->nArcs = 2*(nR/2)+1;
			a->arcs = _p3dArcsCreate(a->nArcs, nNodes,
					distance, 
					a->nodes[a->n-1].x,
					a->nodes[a->n-1].y,
					a->nodes[a->n-1].theta,
					rMin,  depth-1, local_depth+1);

			if (a->arcs == NULL) {
				p3dArcsDestroy(arcs, 2*nR+1);
				errno = ENOMEM;
				return NULL;
			}

			for(j=0;j<a->nArcs;j++) {
				a->arcs[j].parent = a;
				a->arcs[j].nodes[0].prev = &a->nodes[a->n-1].conf;
			}
		}
	}

	return arcs;
}

P3DArc *
p3dArcsCreate(int nArcs, int nNodes, double distance, double x, double y,
		double theta, double rMin, int depth)
		
{
	return _p3dArcsCreate(nArcs, nNodes, distance, x, y, theta, rMin, depth, 1);
}


/*
 * p3dArcsDestroy -------------------------------------------------------
 *
 */

void
p3dArcsDestroy(P3DArc *arcs, int n)
{
	P3DArc *a;
	int i;

	if (arcs == NULL) return;
	if (arcs->nArcs != 0) p3dArcsDestroy(arcs->arcs, arcs->nArcs);

	for (i = 0, a = arcs; i < n; i++, a++) {
		if (a->nodes != NULL) 
			free(a->nodes);
	}
	free(arcs);
}

/*
 * p3dArcComputeNodePosition -----------------------------------------------------
 *
 * Recompute the position of the nodes (in the robot frame) using dMax = stop_distance
 */

void
p3dArcComputeNodePosition(P3DArc *arcs, int n, double x, double y, double theta,
						  double stop_distance)
{
	int i, j;
	P3DArc *a;
	double ds;
	ds = stop_distance / arcs->n;
	if (fabs(ds - arcs->ds) < 1e-5) // same stop_distance, no need to recompute
		return;

	for (i = 0, a = arcs; i<n; i++, a++) {
		a->ds = ds; // a->n is constant so is a->ds
		for (j = 0; j < a->n; j++) {
			p3dGetArcPos(a->r, j*a->ds + a->ds, x, y, theta,
					&a->nodes[j].x, &a->nodes[j].y, &a->nodes[j].theta);

		}

		if (a->nArcs != 0) {
			p3dArcComputeNodePosition(a->arcs, a->nArcs, 
									  a->nodes[a->n - 1].x,
									  a->nodes[a->n - 1].y,
									  a->nodes[a->n - 1].theta,
									  stop_distance);
		}

	}
}


/*
 * p3dArcsResetFlags  --------------------------------------------------
 *
 * Reset any flags (cost, current, first_unknow,  on all arcs
 */
void
p3dArcsResetFlags(P3DArc *arcs, int n)
{
	int i;
	P3DArc *a;

	for(i=0,a=arcs; i<n; i++, a++) {
		a->cost = P3D_UNDEF_COST;
		a->current = 0;
		a->first_unknow = 0;
		a->is_elected = 0;

		if (a->nArcs != 0) {
			p3dArcsResetFlags(a->arcs, a->nArcs);
		}
	}
}


/*
 * p3dArcsReset ---------------------------------------------------------
 *
 * Reset arcs : costs are marked as invalid and remaining distance to
 * goal is computed.
 */

void
p3dArcsReset(P3DArc *arcs, int n,
	     double x, double y, double theta,
	     double gx, double gy, double goalTol)
{
	P3DArc *a;
	int i, j;
	double d, dg;
	double cost = cos(theta), sint = sin(theta);
	double xn, yn, thetan;

	double localMin;
	int localMinNumber;

	for(i=0,a=arcs; i<n; i++, a++) {
		a->cost = P3D_UNDEF_COST;
		a->current = 0;
		a->first_unknow = 0;
		a->is_elected = 0;

		if (a->nArcs != 0) {
			p3dArcsReset(a->arcs, a->nArcs, x, y, theta, gx, gy, goalTol);

			/* get remaining distance from child arcs (min) */
			d = a->arcs[0].nodes[0].remain;
			for(j=1; j<a->nArcs; j++) d = MIN(d, a->arcs[j].nodes[0].remain);
			d += a->ds;
		} else {
			/* end of arc: compute distance to goal */
			xn = a->nodes[a->n-1].x*cost - a->nodes[a->n-1].y*sint + x;
			yn = a->nodes[a->n-1].x*sint + a->nodes[a->n-1].y*cost + y;
			thetan = a->nodes[a->n-1].theta + theta;

			d = p3dDubbins(gx, gy, xn, yn, thetan, -1.0/arcs[0].r);
			/* ???? 	 d -= goalTol; */
			if (d < 0.0) d = 0.0;
		}

		localMin = goalTol + 1.; /* > goalTol */
		localMinNumber = -1;

		for(j=a->n-1; j>=0; j--, d+=a->ds) {
			/* must check if that node is too close to goal */
			xn = a->nodes[j].x*cost - a->nodes[j].y*sint + x;
			yn = a->nodes[j].x*sint + a->nodes[j].y*cost + y;
			dg = sqrt(SQ(xn-gx) + SQ(yn-gy));
			if (dg < goalTol) {
				d = dg;
				if (dg < localMin) {
					localMin = dg;
					localMinNumber = j;
				}
			}

			a->nodes[j].conf.q = -1.0;
			a->nodes[j].cost = P3D_UNDEF_COST;
			a->nodes[j].remain = d;
			a->nodes[j].isLocalMin = 0;
		}

		if (localMinNumber >= 0) {
			a->nodes[localMinNumber].isLocalMin = 1;
		}
	}
}

static void
dump_arc(P3DArc *arcs, T3D* mainToOrigin, int n, 
		P3DArcDump * dumparcs, int * index)
{
	P3DArc *a;
	P3DArcDump *dump;
	int i, j;
	double cost  = cos(mainToOrigin->euler.euler[0]);
	double sint  = sin(mainToOrigin->euler.euler[0]);

	for(i=0,a=arcs; i<n; i++, a++) {
		dump = dumparcs + (*index);
		dump->r = a->r;
		dump->cost = a->cost;
		dump->cost_balanced = a->cost_balanced;
		dump->depth = a->depth;
		dump->nArcs = a->nArcs;
		dump->nNodes = a->n;
		dump->current = a->current;
		dump->first_unknow = a->first_unknow;
		dump->is_elected = a->is_elected;
		for (j = 0; j < a->n; j++)
		{
			P3DNode node = a->nodes[j];
			P3DNodeDump *node_dump = dump->nodes+j;

			node_dump->x = node.x * cost - node.y * sint + mainToOrigin->euler.euler[3];
			node_dump->y = node.x * sint + node.y * cost + mainToOrigin->euler.euler[4];
			node_dump->z = node.conf.z;
			node_dump->yaw = node.conf.theta;
			node_dump->roll = node.conf.roll;
			node_dump->pitch = node.conf.pitch;
			node_dump->pUnk = node.conf.pUnk;
			node_dump->cost = node.cost;
			node_dump->remain = node.remain;
		}
		
		(*index)++;
		if (a->nArcs != 0)
			dump_arc(a->arcs, mainToOrigin, a->nArcs, dumparcs, index);
					 
	}
}

int	
p3dDumpP3dArcs(P3DArc *arcs, T3D* mainToOrigin, int n, P3DArcDump * dumpArc)
{
	int i = 0;

	dump_arc(arcs, mainToOrigin, 2*(n/2) + 1, dumpArc, &i);
	dumpArc->nArcsTotal = i;
	return 0;
}
