/************************************************************
 **
 ** Functions for computation of a difficulty map
 ** on the basis of the result of robot placement on a DEM
 **
 ************************************************************/

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>
/* for memset */
#include <string.h>

#include <time.h>
#include <sys/time.h>

#include "libp3d.h"
#include "p3dInt.h"

#define BODYTEST 0

#define PESSIMISTIC 0

/* put 1 if want Time stats */
#define TIME 1

/*-- Local functions ---------*/
double gauss (double    x,
	      double    m, 
	      double    sigma );

double p3d_ngauss (double    x,
		   double    m, 
		   double    sigma );

double getDifmapDifficultySample(DIF_MAP* difmap,
				 int  i,
				 int  j)
{
  return( (difmap->cells+i*difmap->nbCols+j)->difficulty );
}

double getDifmapProbaSample(DIF_MAP* difmap,
			    int  i,
			    int  j,
			    int  k)  /* Which proba is wanted */
{
  return( (difmap->cells+i*difmap->nbCols+j)->probas[k] );
}



/**********************************************************************
 *
 * Creation and allocation of a DIF_MAP
 * 
 *  WARNING: The difficulty array is EMPTY at that point
 *
 **********************************************************************/

DIF_MAP*	
create_difmap ()
{

  DIF_MAP*   difmap;
  /* int        i,j; */
  
  difmap = (DIF_MAP*) malloc (sizeof(DIF_MAP));
  
  if (difmap == NULL) {
	fprintf(stderr, "%s : Can't allocate memory for difmap\n", __func__); 
    return (NULL);
  }

  /* Initializes parameters */
  difmap->nbLines = 0;
  difmap->nbCols = 0;
  
  difmap->xScale = 0.0;
  difmap->yScale = 0.0;

  difmap->xOrigin = 0.0;
  difmap->yOrigin = 0.0;
  
  /** Init the map mainToOrigin as a T3DEuler **/
  difmap->mainToOrigin = t3dNew(T3D_EULER, T3D_ALLOW_CONVERSION);
  
  difmap->cells = NULL;

  return(difmap);

}


/**********************************************************************
 *
 * Print Info about a DIF_MAP
 *
 **********************************************************************/

void
infoDifmap(DIF_MAP*  map)
{
  printf("-------- DifMap Infos --------------------------------\n");
  printf("map->nbLines    = %d\n", map->nbLines);
  printf("map->nbCols     = %d\n", map->nbCols);
  printf("map->xOrigin    = %f\n", map->xOrigin);
  printf("map->yOrigin    = %f\n", map->yOrigin);
  printf("map->scale      = %f, %f\n",  map->xScale, map->yScale);

  printf("Size            = %f m x %f m\n",  map->xScale*map->nbLines, 
	 map->yScale*map->nbCols);
}






/**********************************************************************
 *
 * This function sets the nb of Lines and columns of an existing DIF_MAP
 *  according to the input parameters
 * and then makes the difficulty array allocation accordingly
 *
 **********************************************************************/

/* From now, This function should be named difmapAllocateCellsTab for ex. */
int
difmapAllocateDifficulty (DIF_MAP*  difmap,
			  int       nbLines,
			  int       nbCols)
{
  difmap->nbLines = nbLines;
  difmap->nbCols  = nbCols;

  /* Array allocation */
  difmap->cells = (DIFMAP_CELL*) calloc(difmap->nbLines*difmap->nbCols, sizeof(DIFMAP_CELL) );
  if(difmap->cells == NULL) {
	fprintf(stderr, "%s : Can't allocate memory for cells\n", __func__);
	return -1;
  }

  return 0;
}



/**********************************************************************
 *
 * Destruction of a DIF_MAP
 *
 **********************************************************************/

void
destroy_difmap (DIF_MAP*  difMap)
{
  
  if(difMap->mainToOrigin != NULL)
    t3dDelete(difMap->mainToOrigin);

  if(difMap->cells != NULL)
    free(difMap->cells);

  free(difMap);
}



/**********************************************************************
 *
 * Copy of an existing DIF_MAP
 *
 **********************************************************************/

DIF_MAP*	
copy_difmap (DIF_MAP*  difmap)
{

  DIF_MAP*   difmap_copy;
  int        i,j;
  
  if (difmap == NULL) {
    fprintf(stderr, "%s : Can't copy a NULL difmap\n", __func__); 
    return (NULL);
  }

  difmap_copy = (DIF_MAP*) malloc (sizeof(DIF_MAP));
  
  if (difmap_copy == NULL) {
	fprintf(stderr, "%s : Can't allocate memory for difmap_copy\n", __func__);
    return (NULL);
  }

  /* Initializes parameters */
  difmap_copy->nbLines = difmap->nbLines;
  difmap_copy->nbCols  = difmap->nbCols;
  
  difmap_copy->xScale = difmap->xScale;
  difmap_copy->yScale = difmap->yScale;

  difmap_copy->xOrigin = difmap->xOrigin;
  difmap_copy->yOrigin = difmap->yOrigin;
  
  /** Init the map mainToOrigin as a T3DEuler **/
  difmap_copy->mainToOrigin = t3dNew(T3D_EULER, T3D_ALLOW_CONVERSION);
  t3dCopyIn(difmap_copy->mainToOrigin, difmap->mainToOrigin);

  difmap_copy->xScale = difmap->xScale;
  difmap_copy->yScale = difmap->yScale;

  difmap_copy->deltaLines = difmap->deltaLines;
  difmap_copy->deltaCols  = difmap->deltaCols;

  /* Allocate the data array */
  if(difmap->cells != NULL) {
    difmap_copy->cells = (DIFMAP_CELL*) calloc(difmap_copy->nbLines*difmap_copy->nbCols, sizeof(DIFMAP_CELL) );
    if(difmap_copy->cells == NULL) {
	  fprintf(stderr, "%s : Can't allocate memory for difmap_copy->cells\n", __func__);
	  destroy_difmap(difmap_copy);
	  return NULL;
	}
        
    for(i = 0; i < difmap_copy->nbLines*difmap_copy->nbCols; i++) {
      (difmap_copy->cells+i)->nbFused = (difmap->cells+i)->nbFused;
      (difmap_copy->cells+i)->firstDate = (difmap->cells+i)->firstDate;
      (difmap_copy->cells+i)->lastDate = (difmap->cells+i)->lastDate;

      (difmap_copy->cells+i)->difficulty = (difmap->cells+i)->difficulty;
      (difmap_copy->cells+i)->sigma = (difmap->cells+i)->sigma;
      for(j = 0; j < DIFMAP_NB_CLASSES; j++)
	(difmap_copy->cells+i)->probas[j] = (difmap->cells+i)->probas[j];
    }
  }
  
  return(difmap_copy);
}







/**********************************************************************
 *
 * Writing a DIF_MAP on a file
 *
 **********************************************************************/

short
write_difmap(DIF_MAP*  map,
	     char      fileName[1024])
{
  FILE*  file;
  double x,y;
  int    i,j;

  file = fopen(fileName, "w");
  if (file == NULL)
    {
      fprintf(stderr, "difmap: failed to open file %s\n", fileName);
      return 1;
    }

  /* Write the file heading */
  fprintf(file, "nbLines = %d\t nbCols = %d\n", map->nbLines, map->nbCols);
  fprintf(file, "xScale = %f\t yScale = %f\n", map->xScale, map->yScale);
  fprintf(file, "deltaLines = %d\t deltaCols = %d\n", map->deltaLines, map->deltaCols);
  
  fprintf(file, "xOrigin = %f\t yOrigin = %f\n", map->xOrigin, map->yOrigin);
  fprintf(file, "mainToOrigin =");
  for(i=0; i<6; i++)
    fprintf(file, " %lf", map->mainToOrigin->euler.euler[i]);
  fprintf(file, "\n");

  /* print the cells in file */
  for(i = 0; i < map->nbLines; i++)
   {
     x = map->xOrigin + ((double) i) * map->xScale;
     for(j = 0; j < map->nbCols; j++)
       {
	 y = map->yOrigin + ((double) j) * map->yScale;
	 /*** TODO: write the additionnal information contained in a cell ***/
	 fprintf(file, "%d %d   %lf %lf   %lf %lf %lf %lf %lf %lf\n",
 		 i, j, x, y, (map->cells+i*map->nbCols+j)->difficulty,
		 (map->cells+i*map->nbCols+j)->sigma,
		 (map->cells+i*map->nbCols+j)->probas[0],
		 (map->cells+i*map->nbCols+j)->probas[1],
		 (map->cells+i*map->nbCols+j)->probas[2],
		 (map->cells+i*map->nbCols+j)->confidence);
	 
       }
     fprintf(file, "\n");
   }
  
  fclose(file);


 return 0;
}




/**********************************************************************
 *
 * Reading a DIF_MAP wrote in a file using the fct write_difmap
 *
 **********************************************************************/

DIF_MAP*
read_difmap(char      fileName[1024])
{
  FILE*  file;
  double x,y;
  int    i,j;
  int    k,l;  /* unused */
  DIF_MAP*  map;

  double         difficulty, sigma, confidence;
  double         proba0, proba1, proba2;
  DIFMAP_CELL*   cell;

  file = fopen(fileName, "r");
  if (file == NULL)
    {
      fprintf(stderr, "difmap: failed to open file %s\n", fileName);
	  return NULL;
    }

  /* Creation of the difmap structure */
  map = create_difmap();

  /* Read the file heading */
  fscanf(file, "nbLines = %d\t nbCols = %d\n", &(map->nbLines), &(map->nbCols));
  fscanf(file, "xScale = %f\t yScale = %f\n", &(map->xScale), &(map->yScale));
  fscanf(file, "deltaLines = %d\t deltaCols = %d\n", &(map->deltaLines), &(map->deltaCols));
  
  fscanf(file, "xOrigin = %f\t yOrigin = %f\n", &(map->xOrigin), &(map->yOrigin));
  fscanf(file, "mainToOrigin =");
  for(i=0; i<6; i++)
    fscanf(file, " %lf", &(map->mainToOrigin->euler.euler[i]) );
  fscanf(file, "\n");

  /*-- Allocate the data array --*/
  map->cells = (DIFMAP_CELL*) calloc(map->nbLines*map->nbCols, sizeof(DIFMAP_CELL) );
  if(map->cells == NULL) {
	fprintf(stderr, "%s : Unable to alloc memory for map->cells\n", __func__);
	destroy_difmap(map);
	return NULL;
  }

  /* Read the cells in file */
  cell = map->cells;
  for(i = 0; i < map->nbLines; i++)
   {
     for(j = 0; j < map->nbCols; j++)
       {
	 /* fscanf(file, "i=%d j=%d   x=%lf y=%lf   diff=%lf\n", */
/*  		&k, &l, &x, &y, &((map->cells+i*map->nbLines+j)->difficulty) ); */
	 fscanf(file, "%d %d   %lf %lf   %lf %lf %lf %lf %lf %lf\n", 
		&k, &l, &x, &y, &difficulty, &sigma, &proba0, &proba1, &proba2, &confidence );
	 //	 cell->difficulty = difficulty;
	 (map->cells+i*map->nbCols+j)->difficulty = difficulty;
	 (map->cells+i*map->nbCols+j)->sigma      = sigma;
	 (map->cells+i*map->nbCols+j)->probas[0] = proba0;
	 (map->cells+i*map->nbCols+j)->probas[1] = proba1;
	 (map->cells+i*map->nbCols+j)->probas[2] = proba2;
	 (map->cells+i*map->nbCols+j)->confidence = confidence;	 
	 cell++;
       }
     fscanf(file, "\n");
   }
  
  fclose(file);


  return(map);
}




/*************************************************************/
/** Functions for the computation of "Difficulty Maps" **/
/** (i.e. qualitative maps with evaluation of difficulty given by robot placement function) **/


/*
 * p3dEvalAngleDiff ---------------------------------------------
 *
 * Evaluates an angle in term of dangerosity (for difficulty maps)
 * Returns a value in [0;1], 0.0 beeing the less dangerous and 1.0
 * indicating impossibility or "maximum" difficulty.
 *
 * Current strategy is to compute d^2 = ( fabs(angle)/max )^2
 * others could be considered
 */

static double
p3dEvalAngleDiff(double angle, double max)
{
   double d = fabs(angle)/max;
   
   return SQ(d);
}





/*
 * p3dEvalConfigDiffDala ---------------------------------------------
 *
 * p3dEvalConfigDiff for Dala
 *
 * Evaluate a configuration in term of dangerosity for difficulty maps
 * Returns a value in [0;1], 0.0 beeing the less dangerous and 1.0
 * indicating impossibility or "maximum" difficulty.
 * Current strategy is to take the max dangerosity for each angle. Other
 * strategy could be applied f.i. compute a distance.
 */


double
p3dEvalConfigDiffDala(P3DRobotGeometry *geom,
		      P3DRobotConfig *config)
{
  double alpha; /* Difference fRoll-bRoll */
  double d; /* Difficulty computed */

  /* if configuration was rejected, set max dangerosity: d=1.0 */
  if(config->q < 0.0)
    return 1.0;

   /* check absolute bounds */
   if (fabs(config->roll) >= geom->rollMax ||
       fabs(config->pitch) >= geom->pitchMax )
      return 1.0;

   /* Test positions relatives des essieux compatible (fRoll~=bRoll) */
   /* fRoll-bRoll */
   alpha = config->alpha1 - config->alpha2;
   if( fabs(alpha) >= geom->alphaMax )
     return 1.0;

#if BODYTEST
   /* Check ground clearance */
   if(config->dzBodyGround <= 0.0)
     return 1.0;
#endif

   /* unknown terrain ? */
   if (config->pUnk > geom->unkMax) {
/*       d = 1.0 + 2*(config->pUnk - geom->unkMax)/(1.0 - geom->unkMax); */
     /* Put some negative value to indicate there is no info */
     d = -1.0;
      return d;
   }

   /* absolute danger */
   d =        p3dEvalAngleDiff(config->roll, geom->rollMax);
   d = MAX(d, p3dEvalAngleDiff(config->pitch, geom->pitchMax) );
/*    d = MAX(d, p3dEvalAngleDiff(config->alpha1, geom->alphaMax) ); */
   d = MAX(d, p3dEvalAngleDiff(alpha, geom->alphaMax) );

#if BODYTEST
   d = MAX(d, SQ( (geom->groundClear-config->dzBodyGround) / geom->groundClear) );
#endif



   return d;
}


/*
 * p3dEvalConfigSigmaDala ---------------------------------------------
 *
 * p3dEvalConfigSigma for Dala
 *
 * Evaluates a Sigma associated to the difficulty for that config
 */
double
p3dEvalConfigSigmaDala(P3DRobotGeometry *geom,
		       P3DRobotConfig *config)
{
  double sigmad;  /* sigma for difficulty */

  /******** The sigma associated to difficulty **********/
  sigmad = config->sigmaRoll/geom->rollMax;
  sigmad = MAX(sigmad, config->sigmaPitch/geom->pitchMax );
#if BODYTEST
  sigmad = MAX(sigmad, config->sigmaz/geom/groundClear);
#endif   
  
   return sigmad;
}


/*
 * p3dDifMap -----------------------------------------------------------
 *
 * Computes a difMap from a P3D_MAP and a robot geometry
 *  NB: dfMap and p3dMap have the SAME geometric size
 *
 */

int
p3dDifMap(DIF_MAP  *difMap,   /* output map */
	  P3D_MAP  *p3dMap,    /* input map */
	  int       delta,    /* difmap sampling rate wrt the p3dMap */
	  P3DRobot *robot, P3DRobotGeometry *geom)  /* info about the robot */
{
  /* Robot's info */
  double wheelRadius;
  double wheelWidth;
  double halfAxleLength;
  double robotMaxLength;
  /* info about map covering */
  double delta_theta;       /* theta increment */

  int    nbExtLines, nbExtCols; /* Nb of ext. lines/cols (robot width) */

  P3DRobotConfig  config;
  double  x,y,theta;   /* robot position in map */
  double difficulty, d;   /* difficulty associated to a robot placement */
  double sigmaDiff, sigmad;    /* Sigma associated to the difficulty */
  int  i,j;

#ifdef TIME
  /*---- Time ----*/
  struct timespec tp;
  unsigned long msec;
  double timesec0;  /* Initial Time in seconds */
  double timesec1;  /* Final Time in seconds */
  /*--------------*/
#endif

  /* Check robot and geom are not empty */
  if(robot == NULL || geom == NULL)
    {
      fprintf(stderr, "p3dDifMap: Error: Robot and/or Geom were not created\n");
      return(-1);
    }

  /*----- TODO: Check info from p3dMap correspond in difMap -----*/
  difMap->xOrigin = p3dMap->xOrigin;
  difMap->yOrigin = p3dMap->yOrigin;


  /*-- Robot Geometry useful info --*/
  wheelRadius = geom->rad[0];
  wheelWidth  = geom->len[0];
  halfAxleLength = geom->len[2] + geom->len[1];

  /* Compute increment between two succesive positions on which the robot placement is to be done */
  /* We use a common delta to make a square map */
  /* corresponding lines/cols increments (wrt p3dMap) */
  difMap->deltaLines = delta;
  difMap->deltaCols  = delta;

  /* define size and scale of config useful part of the difmap */
  difMap->xScale  = ((float) difMap->deltaLines) * p3dMap->scale;
  difMap->yScale  = ((float) difMap->deltaCols ) * p3dMap->scale;
  difMap->nbLines = p3dMap->nbLines / difMap->deltaLines;
  difMap->nbCols  = p3dMap->nbCols  / difMap->deltaCols;


  if (difMap->cells != NULL) {
	  free(difMap->cells);
  }
  /* Allocation of the data array */
  difMap->cells = (DIFMAP_CELL*) calloc(difMap->nbLines*difMap->nbCols, sizeof(DIFMAP_CELL) );
    if(difMap->cells == NULL) {
	  fprintf(stderr, "%s : Can't allocate memory for difmap->cells\n", __func__);
	  return -1;
	}

  /* theta increment between two succesive positions on which the robot placement is to be done */
  /* Define the number of robot's orientations that will be tested */
  /* delta_theta should be chosed so that 2 succesive pos. overlap a little */
  delta_theta = atan(wheelRadius/halfAxleLength);
  /* we should check that wheelRadius < halfAxleLength */
  printf("p3dDifMap: delta_theta=%lf\n", delta_theta);
  
  /* first robot position (w.r.t. p3d map) */
  x = 0.0;
  y = 0.0;
  theta = 0.0;
  difficulty = -1.0;

  /*-----------------------------------------------------------------*/
  /* First thing: borders of the map should be considered as Unknown */
  /*  1. Compute the width and length concerned: robot width */
  /* Nb of line/col cells concerned */
  robotMaxLength = MAX(2.0*halfAxleLength,geom->bodyLength);
  nbExtLines = (int) (robotMaxLength / difMap->xScale) +1;
  nbExtCols  = (int) (robotMaxLength / difMap->yScale) +1;
  difMap->nbBorderLines = nbExtLines;
  difMap->nbBorderCols  = nbExtCols;
  printf("DifMap: halfAxleLength=%lf, nbExtLines=%d, nbExtCols=%d\n", 
	 halfAxleLength, nbExtLines, nbExtCols);
  /* Bottom */
  for(i=0; i < nbExtLines; i++) {
    for(j=0; j < difMap->nbCols; j++)
      (difMap->cells+i*difMap->nbCols+j)->difficulty = -1.0;
  }
  /* High */
  for(i=difMap->nbLines-nbExtLines; i<difMap->nbLines; i++) {
    for(j=0; j < difMap->nbCols; j++)
      (difMap->cells+i*difMap->nbCols+j)->difficulty = -1.0;
  }
  /* Left */
  for(j=0; j < nbExtCols; j++) {
    for(i=0; i< difMap->nbLines; i++)
      (difMap->cells+i*difMap->nbCols+j)->difficulty = -1.0;
  }
  /* Right */
  for(j=difMap->nbCols-nbExtCols; j<difMap->nbCols; j++) {
    for(i=0; i < difMap->nbLines; i++)
      (difMap->cells+i*difMap->nbCols+j)->difficulty = -1.0;
  }
  /*-----------------------------------------------------------------*/
  
#ifdef TIME
  /*----- Start chrono -----*/
  clock_gettime(CLOCK_REALTIME, &tp);
  msec = (tp.tv_nsec / 1000000) + (tp.tv_sec * 1000);
  timesec0  = ((double)(msec))/1000;
#endif

  /* First x */
  x = (double) nbExtLines * difMap->xScale;
  /* Now that params are ready, test placements all over the map */
  /* Except for the borders that have just been excluded */
  for(i=nbExtLines; i< difMap->nbLines-nbExtLines; i++)   /* x (lines) loop */
    {
      /* y = 0.0; */
      y = (double) nbExtCols * difMap->yScale;
      for(j=nbExtCols; j< difMap->nbCols-nbExtCols; j++)  /* y (cols) loop */
	{
	  theta = 0.0;
	  difficulty = -1.0;
	  sigmaDiff  = 0.0;
	  /* Test on present location with different orientations */
	  do {
	    if( p3dPlaceRobotDala(robot, geom, x,y,theta, p3dMap, &config) < 0.0 )
	      {
		difficulty = 1.0;  /* config rejected => difficulty max */	    
		break;   /* No use to go on, so exit the theta loop */
	      }
	    else     /* Compute difficulty on that location */
	      {
		d = p3dEvalConfigDiffDala(geom, &config);
		difficulty = MAX(difficulty, d);
		/* Should we check the max? or the average? or sth else? */

		/** NEW: Associated sigma **/
		sigmad = p3dEvalConfigSigmaDala(geom, &config);
		sigmaDiff = MAX(sigmaDiff,sigmad);
	      }
	    
	    theta += delta_theta;

	  } while (theta < M_PI);
	  
	  /* record difficulty computed */
	  (difMap->cells+i*difMap->nbCols+j)->difficulty = difficulty;

	  /* and associated sigma */
	  (difMap->cells+i*difMap->nbCols+j)->sigma = sigmaDiff;
	  	  
	  y += difMap->yScale;
	}

      x += difMap->xScale;
    }
  
#ifdef TIME
  /*----- End chrono -----*/
  clock_gettime(CLOCK_REALTIME, &tp);
  msec = (tp.tv_nsec / 1000000) + (tp.tv_sec * 1000);
  timesec1  = ((double)(msec))/1000;

  fprintf(stderr, "p3dDifmap : %f s\n", timesec1-timesec0);
#endif

  return 0;
}




/**************************************************************************************
 *
 * Calculates the result of the Gauss function
 * Returns this result
 *
 **************************************************************************************/

double
gauss (double    x,
       double    m,        /* m = average (moyenne ou Esperance) */
       double    sigma )    /* sigma^2 = variance  */
  
{
  return ( (1/sqrt(2.0*M_PI*sigma*sigma)) * exp(-(x-m)*(x-m)/(2.0*sigma*sigma)) );
}



/**************************************************************************************
 *
 * Calculates the result of the Gauss function (normalized: ie max = 1.0)
 * Returns this result
 *
 **************************************************************************************/

double
p3d_ngauss (double    x,
	    double    m,        /* m = average (moyenne ou Esperance) */
	    double    sigma )    /* sigma^2 = variance  */
  
{
  return ( exp(-(x-m)*(x-m)/(2.0*sigma*sigma)) );
}





/**************************************************************************************
 *
 * Computes the "distance" (number of pixels) between two pixels (k,l) and (i,j) 
 *  (to be used for a difmap, but could be used for any image)
 *
 **************************************************************************************/

int
pixDist (int  k,
	 int  l,
	 int  i,
	 int  j)
  
{
  return( MAX( fabs(k-i), fabs(l-j) ) );
}




/**********************************************************************
 *
 * Application of a Gaussian filter to a difmap
 *  Returns the filtered difmap
 *
 **********************************************************************/

DIF_MAP*
filterDifmap (DIF_MAP*  difmap,
	      int      numPixSigma ) /* Num of pixels that will correspond to 2*sigma in the Gaussian used to compute the weights */
{
  int   i,j;
  int   k,l;
  double* weights;  /* weights array */
  double  weightsSum=0.0; 
  double sigma;
  int width=numPixSigma;  /* width of filter (number of pixels) */
  DIF_MAP* filtered_difmap;

  double  diff;
  int     nbLines,nbCols;


  nbLines = difmap->nbLines;
  nbCols  = difmap->nbCols;
  
  if(numPixSigma < 1) {
    fprintf(stderr, "filterDifmap: Error: numPixSigma < 1\n");
    return NULL;
  }

  /* Check numPixSigma vlaue */
  sigma = ((double) width)/2.0;

  /* Allocation of weights array */
  if ( (weights = calloc ((width+1), sizeof (double)) ) == NULL ) {
	 fprintf(stderr, "%s : Can't allocate memory for weights\n", __func__);
	 return NULL;
  }

  /* copy the original difmap as a "basis" for the filtered one */
  filtered_difmap = copy_difmap(difmap);

  /* Calculation of the "weights" determined by the Gaussian function */
  for (i = 0; i <= width; i++)
    weights[i] = gauss( (double) i, 2*sigma, sigma );  /* == gauss(i-3*sigma,0,sigma) */


  /* close to the borders: no way to bother, just copy the difficulty */
  /*  => nothing to be done as it is already in the copy */
  for(i = 0; i < numPixSigma; i++) {
    for(j = 0; j < difmap->nbCols; j++) {
    }
  }
    
  /*-- Compute weightsSum for the average calculation --*/
  for(k = -numPixSigma; k <= numPixSigma; k++) {
    for(l = -numPixSigma; l <= numPixSigma; l++) {
	  /* The weight ot be applied depends on the pixel distance to pixel(i,j) */
      weightsSum += weights[pixDist(k,l,0,0)];
    }
  }
  

  /* Complete part apply the filter using the pixels around */
  for(i = numPixSigma; i < difmap->nbLines-numPixSigma; i++) {
      for(j = numPixSigma; j < difmap->nbCols-numPixSigma; j++) {

	diff = getDifmapDifficultySample(difmap, i, j);

	/** Si difficulty =1 ou -1 juste re-affecter la difficulty ?? ****/
	if( (diff >= 1.0) || (diff < 0.0) ) {
	  /* Nothing special to be done */
	}
	else {
	  (filtered_difmap->cells+i*nbCols+j)->difficulty = 0.0;
#if PESSIMISTIC
	  /*-- In this case a pixel is only influenced by pixels with a higher difficulty --*/
	  weightsSum = 0.0;
	  for(k = i-numPixSigma; k <= i+numPixSigma; k++) {
	    for(l = j-numPixSigma; l <= j+numPixSigma; l++) {
	      if(getDifmapDifficultySample(difmap,k,l) > getDifmapDifficultySample(difmap,i,j)) {
		/* The weight to be applied depends on the pixel distance to pixel(i,j) */
		weight = weights[pixDist(k,l,i,j)];
		(filtered_difmap->cells+i*nbCols+j)->difficulty += weight*getDifmapDifficultySample(difmap,k,l);
		weightsSum += weight;
	      }
	    }
	  }  
#else
	  for(k = i-numPixSigma; k <= i+numPixSigma; k++) {
	    for(l = j-numPixSigma; l <= j+numPixSigma; l++) {
		/* The weight to be applied depends on the pixel distance to pixel(i,j) */
		(filtered_difmap->cells+i*nbCols+j)->difficulty += weights[pixDist(k,l,i,j)]*getDifmapDifficultySample(difmap,k,l);
	    }
	  }
#endif
	  
	  /* Final weighted average value */
	  (filtered_difmap->cells+i*nbCols+j)->difficulty = getDifmapDifficultySample(filtered_difmap,i,j)/weightsSum;
	}
      }
  }
      
      free(weights);
      
  return(filtered_difmap);
}



/**********************************************************************
 *
 * Application of a Gaussian filter to a difmap
 *  Returns the filtered difmap
 *
 **********************************************************************/

int
fuseDifmap (DIF_MAP*  globalDifmap,  /* fused difmap: input and output */
	    DIF_MAP*  localDifmap)   /* local difmap providing new information to add in the global difmap */
{
  double diff;
  int  i0, j0; /* index of localDifmap center in the global array */
  int  i, j, k;

  /* First: check that the scales of both difmap is the same */
  if( (globalDifmap->xScale != localDifmap->xScale) ||
      (globalDifmap->yScale != localDifmap->yScale) ) {
    fprintf(stderr, "fuseDifmap: Error: local and global difmap scales are different !\n");
    return -1;
    }

  /* Second: check that globalDifmap is larger than localDifmap */
  if( (globalDifmap->nbLines < localDifmap->nbLines) ||
      (globalDifmap->nbCols  < localDifmap->nbCols) ) {
    fprintf(stderr, "fuseDifmap: Error: the local difmap is larger than the global difmap !\n");
	return -1;
    }

  /* find the position of local difmap origin in the global one */
  i0 = (int) ((localDifmap->xOrigin-globalDifmap->xOrigin)/globalDifmap->xScale);
  j0 = (int) ((localDifmap->yOrigin-globalDifmap->yOrigin)/globalDifmap->yScale);
  if((i0<0) || (j0<0)) {
    fprintf(stderr, "fuseDifmap: Error: local difmap Origin is out of the global difmap !\n");
	return -1;
  }

  /* Check that the whole local map is inside the global one */
  if( (i0+localDifmap->nbLines)>globalDifmap->nbLines ||
      (i0+localDifmap->nbCols)>globalDifmap->nbCols ) {
    fprintf(stderr, "fuseDifmap: Error: local difmap goes out of global difmap !\n");
	return -1;
  }

  /* Fuse or copy the information of the local map into the global one */
  /* NB: for the moment, the old data is just replaced by the data from local difmap, IF not unknown (ie diff = -1) */
  for(i=0; i<localDifmap->nbLines; i++) {
    for(j=0; j<localDifmap->nbCols; j++) {
      diff = (localDifmap->cells+i*localDifmap->nbCols+j)->difficulty;
      if(diff >= 0.0) /*ie: only if not unknown, replace  */
	(globalDifmap->cells+(i0+i)*globalDifmap->nbCols+(j0+j))->difficulty
	  = diff;
      /* Also Fuse the probas */
      for(k=0; k<DIFMAP_NB_CLASSES; k++)
	(globalDifmap->cells+(i0+i)*globalDifmap->nbCols+(j0+j))->probas[k] = (localDifmap->cells+i*localDifmap->nbCols+j)->probas[k];
    }
  }

  return 0;
}



/**********************************************************************
 *
 * "Probabilization" of a difmap: 
 * ------------------------------
 *  computation of the partial probas of classes flat/rougth/obstacle
 *  according to the difficulty data
 *  Proba distribution for each class is set to a gaussian
 *  Averages are:
 *               - 0.0 for flat terrain
 *               - avRough (input) for rough terrain
 *               - 1.0 for obstacle
 *  The sigmas are input parameters
 *  TODO: provide averages and sigmas inputs as arrays ?
 *
 * !! CAREFUL !! Only works if nb DIFMAP_NB_CLASSES = 3 for now
 *
 **********************************************************************/

int
probaDifmap (DIF_MAP*  difmap,
	     double    avRough,
	     double    sigmaFlat,
	     double    sigmaRough,
	     double    sigmaObstacle)
{
  double  average[DIFMAP_NB_CLASSES];
  double  sigma[DIFMAP_NB_CLASSES];
  double  diff, sigmadiff[DIFMAP_NB_CLASSES];  /* (average) difficulty and associated sigma, "user-weighted" for each classe */
  double  maxProba; /* TEST: max de proba (<= 1.0) */
  double  probaRough;
  DIFMAP_CELL* cell;
  int i,j,k;

  /* Temporary Tests to be check */
  if(DIFMAP_NB_CLASSES != 3) {
    fprintf(stderr, "probaDifmap: ERROR only 3 classes of terrain can be managed for now!\n");
    return -1;
  }

  /*---First, fill the averages and sigma arrays---*/
  average[0] = 0.0;     /* Flat */
  average[1] = avRough; /* Rough */
  average[2] = 1.0;     /* Obstacle */
  sigma[0] = sigmaFlat;   /* Flat */
  sigma[1] = sigmaRough;  /* Rough */
  sigma[2] = sigmaObstacle;  /* Obstacle */


  /** TEST... assumes sigma[0] = sigma[2] ???? **/
/*   maxProba = gauss(average[0],average[0],sigma[0]); */
  maxProba = 1.0;

  /*------ go through the difmap -----------*/
  for(i=0; i<difmap->nbLines; i++) {
    for(j=0; j<difmap->nbCols; j++) {
      cell = &difmap->cells[i*difmap->nbCols+j];
      diff = cell->difficulty;
      if(diff >= 0.0) { /* i.e. cell not unknown */
	for(k=0; k < DIFMAP_NB_CLASSES; k++) {
	  diff = cell->difficulty;
	  /* On est oblige d'affecter un minimum (discretisation) pour le cas ou difficulty = 1.0 car alors le sigma n'est pas calcule => 0.0 */
/* 	  sigmadiff = MAX(cell->sigma, 0.029); */
	  sigmadiff[k] = MAX(cell->sigma, sigma[k]);
	  cell->probas[k]
	    = p3d_ngauss(diff,average[k],sigmadiff[k]);
	  
	} /* end for classes */
      }
      else {
	cell->probas[0] = 0.0;
	cell->probas[1] = 0.0;
	cell->probas[2] = 1.0;
      } /* end if */

      /*** TEST de proba[Rough] = max - ??? */
      probaRough = maxProba - (cell->probas[0]+cell->probas[2]);

      /* Check the proba obtained is in [0,1].... */
      if(probaRough < 0.0) {
	if (probaRough < -0.0001) {
	  printf("probaDifmap: probaRough=%lf is not in [0;1] ! Adjusting...\n",
		 cell->probas[1]);
	}
	probaRough = 0.01;
      }
      else if (probaRough >= 1.0 ) {
	if (probaRough > 1.0001) {
	  printf("probaDifmap: probaRough=%lf is not in [0;1] ! Adjusting...\n",
		 cell->probas[1]);
	}
	probaRough = 0.99;
      }
      cell->probas[1] = probaRough;
    }
  }    
  /**** TEST: Valeur de proba max avec ces parametres ? ***/
#ifdef DEBUG
  printf("max[0] = %lf\n", p3d_ngauss(0.0,0.0,0.03) );
  printf("max[2] = %lf\n", p3d_ngauss(average[2],average[2],sigma[2]) );
#endif  

  return 0;
}




/**********************************************************************
 *
 * Makes a "re-discretisation" of a difmap, to get patches of a specified size
 *
 * returns a difmap containing those patches
 *
 * NB: For now, only works if xScale = yScale, and only makes square patches
 **********************************************************************/

DIF_MAP*
patchDifmap (DIF_MAP*  difmap,  /* input difmap */
	     int       discrRate)  /* Size (in nb of cells) of a patch side */
{
  int   i,j;
  int   k,l;
  int   h;
  int   mi,mj;  /* "Mini" index */
  double diff,maxdiff, sumdiff, averagediff;
  double nbCells;  /* Number of difmap cells in a pdifmap patch */
  double proba[DIFMAP_NB_CLASSES];
  double sum0,max0,max2;
  DIF_MAP*  pdifmap;  /* the output difmap */

  /* Check input difmap is not empty */
  if(difmap->cells == NULL) {
    printf("patchDifmap: ERROR: input difmap is empty\n");
    return NULL;
  }
    
  
  pdifmap = create_difmap();

  pdifmap->nbLines = (int) difmap->nbLines/discrRate;
  pdifmap->nbCols  = (int) difmap->nbCols/discrRate;
  
  /** Allocation cells array **/
  pdifmap->cells = (DIFMAP_CELL*) calloc(pdifmap->nbLines*pdifmap->nbCols, sizeof(DIFMAP_CELL) );
  if(pdifmap->cells == NULL) {
    fprintf(stderr, "%s : Can't allocate memory for pdifmap->cells\n", __func__);
	destroy_difmap(pdifmap);
	return NULL;
  }

  /* New Scale */
  pdifmap->xScale = difmap->xScale * (float) discrRate;
  pdifmap->yScale = difmap->yScale * (float) discrRate;

  /* Same origin */
  pdifmap->xOrigin = difmap->xOrigin;
  pdifmap->yOrigin = difmap->yOrigin;
  pdifmap->mainToOrigin = t3dNew(T3D_EULER, T3D_ALLOW_CONVERSION);
  t3dCopyIn(pdifmap->mainToOrigin, difmap->mainToOrigin);

  /* same deltaLines and deltaCols ?? */

  /* Number of difmap cells in a pdifmap patch */
  nbCells = (double) (discrRate*discrRate);
  
  /* for each pdifmap cell */
  for(i=0; i<pdifmap->nbLines; i++) {
    for(j=0; j<pdifmap->nbCols; j++) {
      maxdiff = 0.0;
      sumdiff = 0.0;
      max2 = 0.0;
      max0 = 0.0;
      sum0 = 0.0;
      
      /* Loop over difmap cells inside */
      for(k=0; k<discrRate; k++) {
	for(l=0; l<discrRate; l++) {
	  /* Mini indexes */
	  mi = discrRate*i+k;
	  mj = discrRate*j+l;
	  
	  diff = getDifmapDifficultySample(difmap,mi,mj);
	  maxdiff = MAX(maxdiff,diff);
	  /* To compute the average */
	  sumdiff += diff;

	  for(h=0; h<DIFMAP_NB_CLASSES; h++)
	    proba[h] = (difmap->cells+mi*difmap->nbCols+mj)->probas[h];
	  max2 = MAX(max2,proba[2]);
	  sum0 += proba[0];
	  max0 = MAX(max0,proba[0]);
	}
      }

      /* The average difficulty value */
      averagediff = sumdiff / nbCells;
      /* The average proba[0] */
      proba[0] = sum0 / nbCells;

      (pdifmap->cells+i*pdifmap->nbCols+j)->difficulty = averagediff; /* maxdiff */
      (pdifmap->cells+i*pdifmap->nbCols+j)->probas[0] = proba[0];
      (pdifmap->cells+i*pdifmap->nbCols+j)->probas[1] = 0.0;  /* Temp. */
      (pdifmap->cells+i*pdifmap->nbCols+j)->probas[2] = max2;

    } /* End for pdifmap Cols */
  }   /* End for pdifmap Lines  */

  return pdifmap;
}





/**********************************************************************
 *
 * Get the partial probas from a difmap
 *  given a mainToOrigin position (x,y)
 *
 *  (+ add the difficulty ??)
 **********************************************************************/
int
getDifmapProbas(DIF_MAP*   difmap,  /* input difmap containing the probas */
		double     x,
		double     y,   /* Position x,y of the robot */
		double     probas[DIFMAP_NB_CLASSES])
{
  int  i,j;  /* coordinates (lines,cols) of the right cell */
  int  k;

  /* find the (i,j) position in difmap corresponding to (x,y) */
  i = (int) ((x-difmap->xOrigin)/difmap->xScale);
  j = (int) ((y-difmap->yOrigin)/difmap->yScale);
  /* Check the position found is inside difmap */
  if((i<0) || (j<0) || (i>=difmap->nbLines) || (j>=difmap->nbCols)) {
    fprintf(stderr, "fuseDifmap: Error: given point (x,y) is out of difmap !\n");
    return -1;
  }

  for(k=0; k<DIFMAP_NB_CLASSES; k++)
    probas[k] = getDifmapProbaSample(difmap, i, j, k);

  return 0;
}




/**********************************************************************
 *
 * Get the difficulty from a difmap (output)
 *  given a mainToOrigin position (x,y)
 *
 **********************************************************************/
double
getDifmapDifficulty(DIF_MAP*   difmap,  /* input difmap containing the probas */
		    double     x,
		    double     y )   /* Position x,y of the robot */
{
  int  i,j;  /* coordinates (lines,cols) of the right cell */

  /* find the (i,j) position in difmap corresponding to (x,y) */
  i = (int) ((x-difmap->xOrigin)/difmap->xScale);
  j = (int) ((y-difmap->yOrigin)/difmap->yScale);
  /* Check the position found is inside difmap */
  if((i<0) || (j<0) || (i>=difmap->nbLines) || (j>=difmap->nbCols)) {
    fprintf(stderr, "fuseDifmap: Error: given point (x,y) is out of difmap !\n");
	return -1.0;
  }

  return( getDifmapDifficultySample(difmap, i,j) );
}
