#if defined(TROOPER)

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>

#include "libp3d.h"
#include "p3dInt.h"
#include "p3dFullImpl.h"

#define DEBUG 0

static int p3dComputeEvo(P3DPlanner *planner, P3D_MAP *map,
		double x, double y, double theta, double alpha, double v,
		P3DPlanSolution *sol);
static void p3dComputeDistanceAlpha(P3DArc * arcs, int nArcs, double r_ref);
static double fDistanceCurvature(double r_ref, double r);
static void  p3dEvoArcPonderateCost(P3DPlanner *planner,  P3DArc* arc, double curv);
static void p3dExtractLeftAndRightArcs(const P3DArc * arc, int n, 
		const P3DArc ** arc_left, const P3DArc ** arc_right, double curv);
static const P3DArc* p3dEvoElectBestArc(const P3DPlanner *planner, const P3DArc *a);

const double epsilon = 1e-4;

int
p3dEvo(P3DPlanner *planner, P3D_MAP *map,
	T3D *mainToOrigin, double alpha, double v,
	P3DPlanSolution * sol)
{
   T3D mainToCenterOfMap;
   int res;
   struct timeval begin, end;
   double diff;
   double x,y,theta;
   double speed;

   gettimeofday(&begin, NULL);

   if (planner->robot == NULL) {
      fputs("p3d: robot not configured\n", stderr);
      return -1;
   }
   if (planner->arcs == NULL) {
      fputs("p3d: arcs not configured\n", stderr);
      return -1;
   }

   p3dComputeRobotPositionInMap(map, 
		   mainToOrigin, &mainToCenterOfMap, 
		   NULL, NULL);

   /* 
	* Compute the position that the robot will approximatvly have
	* when p3dPlan will be finised. For that, use the average time
	* and the previous speed orders
	*/

   x = mainToCenterOfMap.euler.euler[3];
   y = mainToCenterOfMap.euler.euler[4];
   theta = mainToCenterOfMap.euler.euler[0];
   
   p3dComputeRobotMovement(&x, &y, &theta, sol, planner);

   res = p3dComputeEvo(planner, map, x, y, theta, alpha, v, sol);

   speed = max_speed(sol->bestArc, planner);
   if (v > speed) {
	   sol->v = speed;
   } else { 
	   sol->v = v;
   };

   if (res == -1) {
	   sol->v = 0;
	   sol->alpha = alpha;
   }

   gettimeofday(&end, NULL);

   diff = sub_timeval(&begin, &end);
   sol->average_time = ( sol->average_time * sol->ncalls + diff ) / (sol->ncalls+1);
   sol->ncalls++;

   return (fabs(sol->v - v) >= epsilon || fabs(sol->alpha-alpha) >= epsilon);
}

static int
p3dComputeEvo(P3DPlanner *planner, P3D_MAP *map,
		double x, double y, double theta, double alpha, double v,
		P3DPlanSolution *sol)
{
   P3DArc *arcs_alpha, *arc_alpha;	
   const P3DArc* best; 
   double stop_distance;
   double curv = alpha / p3dGetDistanceFixAndMobileAxle(planner);

   p3dResetPlanSolution(sol);

   /* It may be better to compute stop_distance / arcs->max_depth */
   stop_distance = MAX(p3dGetStopDistance(v, planner), planner->dMin);

   /* 
	* Recompute the node position with the new stop_distance 
	* it may be better to only recompute if stop_distance_previous != stop_distance
	* */
   p3dArcComputeNodePosition(planner->arcs, planner->nArcs, 0.0, 0.0, 0.0, stop_distance);

   /* Compute the special arc for alpha */
   /* We can't call p3dArcsCreate for 1 arcs because there is a nice bug. nR
	* will become 0 and leads to a->r = nan*/
   /* Don't forgive it creates 3 Arcs but we are only interested in the last one */
   arcs_alpha = p3dArcsCreate(2, planner->arcs->n, stop_distance, 0.0, 0.0, 0.0,
		   1/curv, 1);

   arc_alpha = (arcs_alpha+2);

   /* Compute the distance of each arc to alpha */
   p3dComputeDistanceAlpha(planner->arcs, planner->nArcs, curv);
   p3dComputeDistanceAlpha(arc_alpha, 1, curv);
	
   /* Compute the cost of each arcs */
   p3dArcComputeCost(planner, map, planner->arcs, planner->nArcs, x, y, theta, 0.0);
   p3dArcComputeCostLimited(planner, map, arc_alpha, 1, x, y, theta, 0.0, 
		   arc_alpha->r - 0.1, arc_alpha->r + 0.1);

   /* Pondate the cost of each arcs */
   p3dEvoArcPonderateCost(planner, arc_alpha, curv);

   /* Elected the best arc */
   best = p3dEvoElectBestArc(planner, arc_alpha);


   if (best == NULL) {
     fprintf(stderr, "evo: best == NULL\n");
	 p3dArcsDestroy(arcs_alpha, 3);
     return -1;
   }

   /* 
	* XXX so crap, best must be modified to show that it is elected, maybe we
	* need to remove this 'not so useful feature'
	*/
   p3dSetPlanSolution(sol, planner, (P3DArc*)best);

   p3dArcsDestroy(arcs_alpha, 3);

   return 0;
}

/*
 * It is clearly an abuse of the sense of node.remain but it let us use the
 * other part of the algorithm inchanged
 */
static void
p3dComputeDistanceAlpha(P3DArc * arcs, int nArcs, double r_ref)
{
	int i;
	P3DArc *a;

	for (i = 0, a = arcs; i < nArcs; i++, a++) {
		double curv_distance = fDistanceCurvature(r_ref, a->r);
		int j;
		for (j = 0; j < a->n; j++) {
			a->nodes[j].remain = curv_distance;
		}
	}
}

static double 
fDistanceCurvature(double r_ref, double r)
{
#if DEBUG
	fprintf(stderr, "Compute distance between %lf and %lf\n", r_ref, r);	
#endif
	double penality = 1.0;
	if (r_ref * r < 0)
		penality = 2.0;
	double res = log2(1.0 + fabs((r - r_ref) * 180 / M_PI));
	return penality * res;
}


/*
 * Not perfectly correct because mostly
 */
static void 
p3dEvoArcPonderateCost(P3DPlanner *planner,  P3DArc* arc, double curv)
{
   const P3DArc* arc_right, *arc_left;
   p3dArcPonderateCost(planner->arcs, planner->nArcs);
   p3dExtractLeftAndRightArcs(planner->arcs, planner->nArcs, &arc_right, &arc_left, curv);
   if (arc_right != NULL && arc_left != NULL) {
	   arc->cost_balanced = fPonderation(arc->cost, arc_right->cost, arc_left->cost);
   } else if (arc_right == NULL) {
	   arc->cost_balanced = fPonderation(arc->cost, arc->n * EASY * arc->ds, arc_left->cost);
   } else {
	   arc->cost_balanced = fPonderation(arc->cost, arc_right->cost, arc->n * EASY * arc->ds);
   }
}

static void 
p3dExtractLeftAndRightArcs(const P3DArc * arc, int n, const P3DArc ** arc_left,
		const P3DArc ** arc_right, double curv)
{
	int i;
	const P3DArc * a;
	const P3DArc * prev = NULL;
	bool found = false;

	*arc_left = NULL;
	*arc_right = NULL;

	a = arc;
	i = 0;

	while (!found && i < n) {

		if (a->r > curv) {
			found = true;
			*arc_left = prev;
			*arc_right = a;
		}

		prev = a;
		a ++;
		i++;
	}

	if (!found) {
		*arc_left = a;
		*arc_right = NULL;
	}
}

static const P3DArc* 
p3dEvoElectBestArc(const P3DPlanner *planner, const P3DArc *a)
{
	const P3DArc * best = p3dElectBestArc(planner->arcs, planner->nArcs);
	if (best == NULL) 
		return NULL;
	(void) a;

	double best_score = p3dFullComputeScore(best);
	double a_score = p3dFullComputeScore(a);

	if (a_score < best_score)
		best = a;

	return best;
}

#endif /* TROOPER */


