/**
 ** David Bonnafous - Anthony Mallet - Fri Jul  6 2001
 ** Copyright (c) 2001 LAAS/CNRS 
 **
 ** $Log: robot.c,v $
 ** Revision 1.4  2007/11/27 10:41:20  adegroot
 ** Clean a bit the code so it can compile with -W -Wall -Wextra -Wwrite-strings -Werror
 **
 ** -pedantic is not supported because we use gnu extension #ident
 **
 ** Revision 1.3  2007/06/04 16:01:25  tpeynot
 ** Correction: nBody must be computed Before the test on its value...
 **
 ** Revision 1.2  2007/01/08 19:55:17  sjoyeux
 ** * sync with the DEM tag (was something else before)
 ** * update to OpenRobot's T3Ds
 **
 ** Revision 1.1.1.1  2006/10/17 18:09:42  sjoyeux
 ** Packaged the P3D module from Lama's calife.
 ** See CHANGES for a list of modifications
 **
 **
 ** Revision 1.8.2.3  2005/04/13 20:27:39  tpeynot
 ** Modifications of protos and structures for body collision test (e.g. ground clearance info)
 **
 ** Revision 1.8.2.2  2005/04/07 20:42:46  tpeynot
 ** Add functions and elements in structures etc... for Dala body collision test
 ** Most of them are commented for the moment
 **
 ** Revision 1.8.2.1  2005/03/24 23:24:37  tpeynot
 ** Just a comment
 **
 ** Revision 1.8  2005/03/17 13:46:54  tpeynot
 ** Add param "bestArc" ro p3dPlan fct to get info about the arc selectes by p3d planner
 ** Add functions adapted to Dala's structure instead of Lama. These functions are used if there's a flag -DATRV in Makefile. Otherwise, the lib will keep working with Lama adapted functions
 **
 ** Revision 1.7  2001/09/13 10:49:44  mallet
 ** Les ponderations des parametres de configuration qui marchent bien
 **
 ** Revision 1.6  2001/07/18 19:46:03  mallet
 ** printfs et finitions
 **
 ** Revision 1.5  2001/07/15 17:44:42  mallet
 ** sync
 **
 ** Revision 1.4  2001/07/14 11:39:09  mallet
 ** premiere version ok
 **
 ** Revision 1.3  2001/07/12 17:38:45  mallet
 ** sync
 **
 ** Revision 1.2  2001/07/12 09:56:00  mallet
 ** plus robuste dans la detection de la fin des iterations du placement (amortissement)
 **
 ** Revision 1.1  2001/07/11 18:05:41  mallet
 ** re-ecriture de p3d
 **
 **/
#ident "$Id: robot.c,v 1.4 2007/11/27 10:41:20 adegroot Exp $"

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>

#include "libp3d.h"
#include "p3dInt.h"

#define DEBUG 0

/*
 * p3dRobotCreate -------------------------------------------------------
 *
 * Create a robot description. This includes a set of bitmaps
 * representing the wheels profile for a range of roll angles.
 * Bitmaps are indexed by roll angle and the precision of placement is
 * the roll step.
 */

P3DRobot *
p3dCreateRobot(P3DRobotGeometry *geom, double scale, double rollStep, double vMin, double vMax,
									   double wMax, double accel, double decel)
{
   P3DRobot *robot;
   int i, j, n;
   int nRoll, nPitch, nBody;
   double  pitchStep;
   double  bodyRollMin;
   double  bodyPitchStep, bodyRollStep;
#if defined(ATRV)
   const char* name = "ATRV";
#elif defined(TROOPER)
   const char* name = "TROOPER";
#elif defined(RMP400)
   const char* name = "RMP400";
#endif

   n = 2 * (geom->rollMax + geom->alphaMax) / rollStep;
   if (n < 1) { errno = EINVAL; return NULL; }

   robot = malloc(sizeof(*robot));
   if (robot == NULL) return NULL;

   robot->damping = 0.75;
   robot->nIter = 10;

   robot->rollMin = -geom->rollMax - geom->alphaMax;
   robot->rollStep = rollStep;
   robot->nProfiles = n;
   robot->lWheel = calloc(n, sizeof(*robot->lWheel));
   if (robot->lWheel == NULL) { 
	  errno = ENOMEM;
	  p3dRobotDestroy(robot);
      return NULL;
   }
   robot->rWheel = calloc(n, sizeof(*robot->rWheel));
   if (robot->rWheel == NULL) {
	  errno = ENOMEM;
	  p3dRobotDestroy(robot);
      return NULL;
   }

#if (defined(ATRV) /* Dala */ || defined(TROOPER) || defined(RMP400))
   /*-------------- ATRV wheels bitmaps --------------------------*/
   /* Creation of wheels profile for ATRV */
   for(i=0; i<n; i++) {
      robot->lWheel[i] =
	 p3dDalaWheelCreate(geom->len, geom->rad,
			    M_PI + i*rollStep + robot->rollMin, scale);
      robot->rWheel[i] =
	 p3dDalaWheelCreate(geom->len, geom->rad,
			    i*rollStep + robot->rollMin, scale);
      if (robot->lWheel[i] == NULL || robot->rWheel[i] == NULL) {
		  errno = ENOMEM;
		  p3dRobotDestroy(robot);
		  return NULL;
      }

      fprintf(stderr, "p3d: created wheel profile %d/%d\r", i, n);
   }

   fprintf(stderr, "p3d: created all %s wheel profiles\t\n", name);

   /*----------------- ATRV body bitmaps --------------------------*/
   bodyRollMin = -geom->rollMax;
   robot->bodyRollMin = bodyRollMin;
   /* Check parameters */
   if(fabs(geom->pitchMax) > M_PI_4)
     fprintf(stderr, "p3d: WARNING: geom->pitchMax=%lf, should be <= PI/4\n", geom->pitchMax);

   /* Parameters for ATRV's body */
   robot->bodyPitchMin = -geom->pitchMax;
   /* Define other step values to compute less body bitmaps */
   pitchStep = rollStep;
   bodyPitchStep = pitchStep*3.0;
   bodyRollStep = rollStep*3.0;
   robot->bodyPitchStep = bodyPitchStep;
   robot->bodyRollStep = bodyRollStep;

   nRoll = 2 * geom->rollMax / bodyRollStep;
   nPitch = 2 * geom->pitchMax / bodyPitchStep;
   nBody = nRoll * nPitch;
   if (nRoll < 1 || nBody < 1) { errno = EINVAL; p3dRobotDestroy(robot); return NULL; }
   
   robot->nBodyRoll = nRoll;
   robot->nBodyPitch = nPitch;
   robot->nBodyProfiles = nBody;

   /* Allocations */
   /* for roll indexation */
   robot->body = calloc(nRoll, sizeof(*robot->body));
   if (robot->body == NULL) { 
	  errno = ENOMEM;
	  p3dRobotDestroy(robot);
      return NULL;
   }
   for(i=0; i<nRoll; i++)
     {
       robot->body[i] = calloc(nPitch, sizeof(**robot->body));
       if (robot->body[i] == NULL) { 
		   errno = ENOMEM;
		   p3dRobotDestroy(robot);
		   return NULL;
       }
     }

   /* Creation of body profile for ATRV */
   for(i=0; i<nRoll; i++) {
     for(j=0; j<nPitch; j++) {
       robot->body[i][j] =
	 p3dDalaBodyCreate(geom->bodyLength, geom->bodyWidth, geom->zBody,
			   i*bodyRollStep + bodyRollMin, 
			   j*bodyPitchStep + robot->bodyPitchMin, scale);
     if (robot->body[i][j] == NULL) {
       p3dRobotDestroy(robot);
       fprintf(stderr, "p3d: Failed to create robot->body[%d][%d]\n", i, j);
       errno = ENOMEM;
       return NULL;
       /* stop the body bitmaps creation, as it failed, but return robot with wheel bitmaps */
       return robot;
     }
     }
     fprintf(stderr, "p3d: creation of %s Body profile %d/%d\r", name, i*j, nBody);
   }
   fprintf(stderr, "p3d: created all %s body profiles\t\n", name);
   
#else  /* Lama */

   /*-------------- LAMA wheels bitmaps --------------------------*/
   /* Creation of wheels profile for Lama */
   for(i=0; i<n; i++) {
      robot->lWheel[i] =
	 p3dLamaWheelCreate(geom->len, geom->rad,
			    M_PI + i*rollStep + robot->rollMin, scale);
      robot->rWheel[i] =
	 p3dLamaWheelCreate(geom->len, geom->rad,
			    i*rollStep + robot->rollMin, scale);
      if (robot->lWheel[i] == NULL || robot->rWheel[i] == NULL) {
	 p3dRobotDestroy(robot);
	 errno = ENOMEM;
	 return NULL;
      }

      fprintf(stderr, "p3d: created wheel profile %d/%d\r", i, n);
   }
   fprintf(stderr, "p3d: created all Lama wheel profiles\t\n");
#endif

   robot->vMin = vMin;
   robot->vMax = vMax;
   robot->wMax = wMax;
   robot->accel = accel;
   robot->decel = decel;

   return robot;
}


/*
 * p3dRobotDestroy ------------------------------------------------------
 *
 * Free robot storage
 */

void
p3dRobotDestroy(P3DRobot *robot)
{
  int i, j;

   if (robot == NULL) return;

   if (robot->lWheel != NULL) {
      for(i=0;i<robot->nProfiles;i++) {
	 p3dP3dElementDestroy(robot->lWheel[i]);
	 p3dP3dElementDestroy(robot->rWheel[i]);
      }
      free(robot->lWheel);
      free(robot->rWheel);
   }

#if (defined(ATRV) || defined(TROOPER) || defined(RMP400))   /* Destruction of body */
   if (robot->body != NULL) {
      for(i=0;i<robot->nBodyRoll;i++) {
	if(robot->body[i] != NULL) {
	  for(j=0;j<robot->nBodyPitch;j++) {
	    p3dP3dElementDestroy(robot->body[i][j]);
	  }
	}
	free(robot->body[i]);
      }
      free(robot->body);
   }
#endif

   free(robot);
}


/*
 * p3dEvalConfig --------------------------------------------------------
 *
 * Evaluate a configuration in term of dangerosity.
 * Returns a value in [1;+inf[, 1 beeing the less dangerous or <0.0,
 * indicating impossibility.
 * Current strategy is to take the max dangerosity for each angle. Other
 * strategy could f.i. compute a distance.
 */

static double	p3dEvalAngle(double angle, double max, double coeff);

double
p3dEvalConfig(P3DRobotGeometry *geom,
	      P3DRobotConfig *current, P3DRobotConfig *prev)
{
   double d;

   /* check absolute bounds */
   if (fabs(current->roll) > geom->rollMax ||
       fabs(current->pitch) > geom->pitchMax ||
       fabs(current->alpha1) > geom->alphaMax ||
       fabs(current->alpha2) > geom->alphaMax ||
       fabs(current->beta3-M_PI) > geom->betaMax)
      return -1.0;

   /* check relative bounds */
   if (fabs(current->roll - prev->roll) > geom->rollMax ||
       fabs(current->pitch - prev->pitch) > geom->pitchMax ||
       fabs(current->alpha1 - prev->alpha1) > geom->alphaMax ||
       fabs(current->alpha2 - prev->alpha2) > geom->alphaMax ||
       fabs(current->beta3 - prev->beta3) > geom->betaMax)
      return -1.0;

   /* unknown terrain ? */
   if (current->pUnk > geom->unkMax) {
      d = 1.0 + 2*(current->pUnk - geom->unkMax)/(1.0 - geom->unkMax);
      return d;
   }

   /* absolute danger */
   d =        p3dEvalAngle(current->roll, geom->rollMax, 1.);
   d = MAX(d, p3dEvalAngle(current->pitch, geom->pitchMax, 1.));
   d = MAX(d, p3dEvalAngle(current->alpha1, geom->alphaMax, 2.0));
   d = MAX(d, p3dEvalAngle(current->alpha2, geom->alphaMax, 2.0));
   d = MAX(d, p3dEvalAngle(current->beta3-M_PI, geom->betaMax, 1.));

   /* relative danger */
   d = MAX(d, p3dEvalAngle(current->roll - prev->roll, geom->rollMax, 1.));
   d = MAX(d, p3dEvalAngle(current->pitch - prev->pitch, geom->pitchMax, 1.));
   d = MAX(d, p3dEvalAngle(current->alpha1 - prev->alpha1,
			   geom->alphaMax, 2.0));
   d = MAX(d, p3dEvalAngle(current->alpha2 - prev->alpha2,
			   geom->alphaMax, 2.0));
   d = MAX(d, p3dEvalAngle(current->beta3 - prev->beta3, geom->betaMax, 1.));

   return d;
}

static double
p3dEvalAngle(double angle, double max, double coeff)
{
   double d = fabs(angle)/max;
   
   /* This can be changed to whatever you want */

   if (d > 1.0-1e-5) return 1e10;
   return 1.0 + coeff*(1.0/SQ(1.0-d) - 1.0);
}


/*
 * p3dLamaWheelCreate ---------------------------------------------------
 *
 * Create a 2D bitmap with heights of lama wheel at the given resolution
 * (scale is the size of pixels in meters).
 * For efficiency, several bitmaps are computed for different
 * orientations (roll angles) [this function computes only one bitmap].
 *
 * Lama wheels are made of three parts : a cylinder, a cone and a
 * cylinder. The sizes of this elements are given by lengths[] and
 * radius[]. The reference point is centered on the left side.
 *
 *      ref        ___+--+ ^
 *       v_____+---   |  | |
 *  r[1] |_____|      |  | | r[0]
 *             +---___|  | |
 *                    +--+ v
 *       <----><-----><-->
 *        l[2]   l[1] l[0]
 *
 */

P3DElement *
p3dLamaWheelCreate(double len[3], double rad[2], double roll, double scale)
{
   P3DElement *wheel;
   double cosr = cos(roll), sinr = -sin(roll), tanr = -tan(roll);
   double w, h, l, r, e, x,y,z;
   double loff, roff;
   float *height;
   int i, j;

   /* horizontal projection of "wheel sides" */
   loff = fabs(rad[1]*sinr);
   roff = fabs(rad[0]*sinr);

   /* bitmap size */
   w = loff + (len[2]+len[1]+len[0])*fabs(cosr) + roff;
   w = MAX(2*loff + len[2]*fabs(cosr), w);
   w = MAX(2*roff + len[0]*fabs(cosr), w);

   h = 2*MAX(rad[0], rad[1]);

   /* create storage */
   wheel = malloc(sizeof(*wheel));
   if (wheel == NULL) { errno = ENOMEM; return NULL; }

   wheel->w = ceil(w/scale);
   wheel->h = ceil(h/scale);
   if (wheel->w < 1 || wheel->h < 1) { 
      errno = EINVAL; free(wheel);
      return NULL;
   }

   wheel->height = malloc(wheel->w * wheel->h * sizeof(*wheel->height));
   if (wheel->height == NULL) { errno = ENOMEM; free(wheel); return NULL; }

   /* offset of ref point */
   loff = MAX(loff, loff-len[2]*cosr);
   loff = MAX(loff, roff-(len[2]+len[1])*cosr);
   loff = MAX(loff, roff-(len[2]+len[1]+len[0])*cosr);

   wheel->scale = scale;
   wheel->x = loff;
   wheel->y = h/2;

   /* fill height map */
   wheel->nElem = 0;
   height = wheel->height;
   for(i=0; i<wheel->h; i++) {

      /* radial position between [-h/2;h/2] (meters) */
      r = h/2.0*(2.0*i/(wheel->h-1) - 1.0);

      for(j=0; j<wheel->w; j++, height++) {

	 /* horizontal position between [-loff;len[0+1+2]+roff] (meters) */
	 l = (w*j/(wheel->w-1) - loff)/cosr;

	 /* altitude if undef by default */
	 z = P3D_UNDEF_ELEM;

	 /* height of axis */
	 e = l*sinr;

	 /* height of profile */

	 if (fabs(tanr) > 1e-5) { /* tiny roll */
	    /* left side */
	    x = -l/tanr;
	    if (x*x+r*r <= rad[1]*rad[1])
	       z = MIN(z, e+x*SIGN(cosr)*sqrt(1+SQ(tanr)));

	    /* right side */
	    y = len[2]+len[1]+len[0]-l;
	    x = y/tanr;
	    if (x*x+r*r <= rad[0]*rad[0])
	       z = MIN(z, e+x*SIGN(cosr)*sqrt(1+SQ(tanr)));
	 }

	 /* left cylinder */
	 if (r<=rad[1]) {
	    x = sqrt(rad[1]*rad[1]-r*r);
	    if (x*tanr+l >= 0 && x*tanr+l <= len[2])
	       z = MIN(z, e+x*SIGN(cosr)*sqrt(1+SQ(tanr)));

	    if (-x*tanr+l >= 0 && -x*tanr+l <= len[2])
	       z = MIN(z, e-x*SIGN(cosr)*sqrt(1+SQ(tanr)));
	 }

	 /* right cylinder */
	 if (r<=rad[0]) {
	    x = sqrt(rad[0]*rad[0]-r*r);
	    if (x*tanr+l >= len[2]+len[1] && 
		x*tanr+l <= len[2]+len[1]+len[0])
	       z = MIN(z, e+x*SIGN(cosr)*sqrt(1+SQ(tanr)));

	    if (-x*tanr+l >= len[2]+len[1] && 
		-x*tanr+l <= len[2]+len[1]+len[0])
	       z = MIN(z, e-x*SIGN(cosr)*sqrt(1+SQ(tanr)));
	 }

	 /* cone */
	 {
	    /* radius r = az+b */
	    double a = (rad[0]-rad[1])/len[1];
	    double b = rad[1]-a*len[2];
	    double rb = a*l+b;

	    /* discriminant de x+y=r */
	    y = (SQ(a*tanr)-1)*SQ(r) + SQ(rb);

	    if (y >= 0) {
	       x = (a*tanr*rb + sqrt(y))/(1-SQ(a*tanr));
	       if (x*tanr+l >= len[2] && 
		   x*tanr+l <= len[2]+len[1])
		  z = MIN(z, e+x*SIGN(cosr)*sqrt(1+SQ(tanr)));

	       x = (a*tanr*rb - sqrt(y))/(1-SQ(a*tanr));
	       if (x*tanr+l >= len[2] && 
		   x*tanr+l <= len[2]+len[1])
		  z = MIN(z, e+x*SIGN(cosr)*sqrt(1+SQ(tanr)));
	    }
	 }

	 /* fixed point height */
	 if (z == P3D_UNDEF_ELEM)
	    *height = P3D_UNDEF_ELEM;
	 else {
	    *height = z;
	    wheel->nElem++;
	 }
      }
   }

   return wheel;
}


/*
 * p3dP3dElementDestroy --------------------------------------------------
 *
 * Free wheel (or body) storage
 */

void
p3dP3dElementDestroy(P3DElement *wheel)
{
   if (wheel == NULL) return;

   if (wheel->height != NULL) free(wheel->height);
   free(wheel);
}




/**********************************************************************/
/********** Thierry **************************************************/



/*
 * p3dDalaWheelCreate ---------------------------------------------------
 *
 * Create a 2D bitmap with heights of Dala wheel at the given resolution
 * (scale is the size of pixels in meters).
 * For efficiency, several bitmaps are computed for different
 * orientations (roll angles) [this function computes only one bitmap].
 *
 * Dala wheels (half-axle) are considered as made of 3 parts: a cylinder corresponding to Dala's body, a small cylinder which is the visible part of the axle, and a third one which is the wheel with its tire.
 * !!!!! Pour se simplifier la vie, on considerera pour le moment que le cylindre 2 est un CONE... (comme sur Lama)
 *
 * The sizes of this element are given by lengths[] and radius[]. The reference point is centered on the left side.
 *                  
 *      ref          +-----+ ^
 *       v______     |     | |
 *       |      |____|     | |
 *  r[1] |      |____|     | | r[0]
 *       |______|    |     | |
 *                   |     | |
 *       <----> <--->+-----+ v
 *        l[2]   l[1]
 *                   <----->
 *                    l[0]
 */

P3DElement *
p3dDalaWheelCreate(double len[3], double rad[2], double roll, double scale)
{
   P3DElement *wheel;
   double cosr = cos(roll), sinr = -sin(roll), tanr = -tan(roll);
   double w, h, l, r, e, x,y,z;
   double loff, roff;
   float *height;
   int i, j;

   /* horizontal projection of "wheel sides" */
   roff = fabs(rad[0]*sinr);

   /* bitmap size */
   w = 2*roff + len[0]*fabs(cosr);

   h = 2*rad[0];

   /* create storage */
   wheel = malloc(sizeof(*wheel));
   if (wheel == NULL) { errno = ENOMEM; return NULL; }

   wheel->w = ceil(w/scale);
   wheel->h = ceil(h/scale);
   if (wheel->w < 1 || wheel->h < 1) { 
      errno = EINVAL; free(wheel);
      return NULL;
   }

   wheel->height = malloc(wheel->w * wheel->h * sizeof(*wheel->height));
   if (wheel->height == NULL) { errno = ENOMEM; free(wheel); return NULL; }


   /* offset of ref point */
   loff = roff-(len[2]+len[1]+len[0])*cosr;

   wheel->scale = scale;
   wheel->x = loff;
   wheel->y = h/2;

   /* fill height map */
   wheel->nElem = 0;
   height = wheel->height;
   for(i=0; i<wheel->h; i++) {

      /* radial position between [-h/2;h/2] (meters) */
      r = h/2.0*(2.0*i/(wheel->h-1) - 1.0);

      for(j=0; j<wheel->w; j++, height++) {

	 /* horizontal position between [-loff;len[0+1+2]+roff] (meters) */
	 l = (w*j/(wheel->w-1) - loff)/cosr;

	 /* altitude if undef by default */
	 z = P3D_UNDEF_ELEM;

	 /* height of axis */
	 e = l*sinr;

	 /* height of profile */

	 if (fabs(tanr) > 1e-5) { /* tiny roll */
	    /* right side */
	    y = len[2]+len[1]+len[0]-l;
	    x = y/tanr;
	    if (x*x+r*r <= rad[0]*rad[0])
	       z = MIN(z, e+x*SIGN(cosr)*sqrt(1+SQ(tanr)));
	 }

	 /* right cylinder */
	 if (r<=rad[0]) {
	    x = sqrt(rad[0]*rad[0]-r*r);
	    if (x*tanr+l >= len[2]+len[1] && 
		x*tanr+l <= len[2]+len[1]+len[0])
	       z = MIN(z, e+x*SIGN(cosr)*sqrt(1+SQ(tanr)));

	    if (-x*tanr+l >= len[2]+len[1] && 
		-x*tanr+l <= len[2]+len[1]+len[0])
	       z = MIN(z, e-x*SIGN(cosr)*sqrt(1+SQ(tanr)));
	 }

	 /* fixed point height */
	 if (z == P3D_UNDEF_ELEM)
	    *height = P3D_UNDEF_ELEM;
	 else {
	    *height = z;
	    wheel->nElem++;
	 }
      }
   }

   return wheel;
}


/*
 * The value returned must be studied a bit more in simulation / real case
 */
static double
p3dEvalAngleDifficulty(double angle, double angleLimits[3])
{
	double absolute_angle = fabs(angle);

	if (absolute_angle > angleLimits[2])
		return -1.0;

	if (absolute_angle < angleLimits[0])
		return EASY; 

	if (absolute_angle < angleLimits[1])
		return MEDIUM; 

	return HARD;
}


/*
 * p3dEvalConfigDala --------------------------------------------------------
 *
 * p3dEvalConfig for Dala
 *
 * Evaluate a configuration in term of dangerosity.
 * Returns a value in [1;+inf[, 1 beeing the less dangerous or <0.0,
 * indicating impossibility.
 * Current strategy is to compute the sum of the difficulty for each angle
 */


#define eval(a,b) \
	d = p3dEvalAngleDifficulty((a), (b)); \
	if ( d < 0 ) \
	  return d; \
	sum += d;	  

/* 
 * Maybe we need to favour the same trajectory using eval_relative
 * instead of eval
 */

#define eval_relative(a,b) \
	if ( a > 1e-5 ) {  \
		d = p3dEvalAngleDifficulty((a), (b)); \
		if ( d < 0 ) \
			return d; \
		sum += d; \
	}

#if defined(DEBUG) && DEBUG
#define DPRINTF(x) printf x
#else
#define DPRINTF(x)
#endif

double
p3dEvalConfigDala(P3DRobotGeometry *geom,
		  P3DRobotConfig *current, P3DRobotConfig *prev)
{
   double alpha, prev_alpha; /* Difference fRoll-bRoll */
   double d;
   double sum = 0;

   /* check absolute bounds */
   if (fabs(current->roll) > geom->rollMax) {
	   DPRINTF(("%s : roll (%lf) > rollMax (%lf)\n", __func__, current->roll, geom->rollMax));
	   return -1.0;
   }

   if (fabs(current->pitch) > geom->pitchMax ) {
	   DPRINTF(("%s : pitch (%lf) > pitchMax (%lf)\n", __func__, current->pitch, geom->pitchMax));
	   return -1.0;
   }

   /* Test positions relatives des essieux compatible (fRoll~=bRoll) */
   /* fRoll-bRoll */
   alpha = current->alpha1 - current->alpha2;
   if( fabs(alpha) > geom->alphaMax ) {
	   DPRINTF(("%s : alpha (%lf) > alphaMax (%lf)\n", __func__, alpha, geom->alphaMax));
	   return -1.0;
   }

   prev_alpha = prev->alpha1 - prev->alpha2;
   /* check relative bounds */
   if (fabs(current->roll - prev->roll) > geom->rollMax) {
	   DPRINTF(("%s : relative roll (%lf) > rollMax (%lf)\n", __func__, 
			   current->roll - prev->roll, geom->rollMax));
	   return -1.0;
   }

   if (fabs(current->pitch - prev->pitch) > geom->pitchMax) {
	   DPRINTF(("%s : relative pitch (%lf) > pitchMax (%lf)\n", __func__, 
			   current->pitch - prev->pitch, geom->pitchMax));
	   return -1.0;
   }

   if (fabs(alpha - prev_alpha) > geom->alphaMax) {
	   DPRINTF(("%s : relative alpha (%lf) >alphahMax (%lf)\n", __func__, 
			   alpha - prev_alpha, geom->alphaMax));
	   return -1.0;
   }

   /* unknown terrain ? */
   if (current->pUnk > geom->unkMax) {
      d = 1.0 + 2*(current->pUnk - geom->unkMax)/(1.0 - geom->unkMax);
      return d;
   }

   /* absolute danger */
   eval(current->roll, geom->rollLimits);
   eval(current->pitch, geom->pitchLimits);
   eval(alpha, geom->alphaLimits);

   /* relative danger */
   /* Maybe we need to ponderate them */
   eval(current->roll - prev->roll, geom->rollLimits);
   eval(current->pitch - prev->pitch, geom->pitchLimits);
   eval(alpha - prev_alpha, geom->alphaLimits);

   /* Here we normalize because we want to have a result in [1..+inf[ */
   sum /= 6; 

   /* TODO: Ajouter ici le danger lie a la proximite de DEM et de la caisse */

   return sum;
}

double
p3dEvalConfigTrooper(P3DRobotGeometry *geom,
	      P3DRobotConfig *current, P3DRobotConfig *prev)
{
   double d;
   double sum = 0.0;

   /* check absolute bounds */
   if (fabs(current->roll) > geom->rollMax ||
       fabs(current->pitch) > geom->pitchMax ||
       fabs(current->alpha1) > geom->alphaMax ||
       fabs(current->alpha2) > geom->alphaMax ||
       fabs(current->beta3-M_PI) > geom->betaMax)
      return -1.0;

   /* check relative bounds */
   if (fabs(current->roll - prev->roll) > geom->rollMax ||
       fabs(current->pitch - prev->pitch) > geom->pitchMax ||
       fabs(current->alpha1 - prev->alpha1) > geom->alphaMax ||
       fabs(current->alpha2 - prev->alpha2) > geom->alphaMax ||
       fabs(current->beta3 - prev->beta3) > geom->betaMax)
      return -1.0;

   /* unknown terrain ? */
   if (current->pUnk > geom->unkMax) {
      //d = 1.0 + 2*(current->pUnk - geom->unkMax)/(1.0 - geom->unkMax);
      //return d;
	  return EASY; /* consider unknow as easy so we will discover it if it is the shortest way */
   }

   /* absolute danger */
   eval(current->roll, geom->rollLimits);
   eval(current->pitch, geom->pitchLimits);
   eval(current->alpha1, geom->alphaLimits);
   eval(current->alpha2, geom->alphaLimits);
   eval(current->beta3 - M_PI, geom->betaLimits);

   /* relative danger */
   eval(current->roll - prev->roll, geom->rollLimits);
   eval(current->pitch - prev->pitch, geom->pitchLimits);
   eval(current->alpha1 - prev->alpha1, geom->alphaLimits);
   eval(current->alpha2 - prev->alpha2, geom->alphaLimits);
   eval(current->beta3 - prev->beta3, geom->betaLimits);

   sum /= 10;

   /* TODO: Ajouter ici le danger lie a la proximite de DEM et de la caisse */

   return sum;
}



/*
 * p3dDalaBodyCreate ---------------------------------------------------
 *
 * Create a 2D bitmap with heights of Dala above body at the given resolution
 * (scale is the size of pixels in meters).
 * For efficiency, several bitmaps are computed for different
 * orientations (roll and pitch angles) 
 * [this function computes only one bitmap].
 *
 * For Dala, this is a simple rectangle to be projected
 * Heights computed are actually the height differences with the z of robot's center
 *
 *                 width (along Robot's y axis)
 *            <----------------->
 *          ^ -------------------
 *          | |A               D|
 * length   | |                 |
 * (x axis) | |        x        |
 *          | |                 |
 *          | |B               C|
 *          v -------------------  
 *
 *  zBody is the delta_z between robot's center z and center of the fond de caisse
 * 
 */

P3DElement *
p3dDalaBodyCreate(double length, double width, double zBody, double roll, double pitch, double scale)
{
  P3DElement *body;
  double w, h, l, r, z;
  float *height;
  int i, j;

  T3D   bodyToRobotRed;  /* Position (orientation) of body wrt. Robot with pitch=roll=0 */
  T3D   pointToBody;   /* Position of a given point wrt. the body */
  T3D    pointToRobotRed;  /* Position of a given point wrt. RobotRed */
  double  xA,yA, xB,yB, xD,yD;
  double  scale_x, scale_y;  /* scale = space between two successive pixels on body plane */

  t3dInit(&pointToRobotRed, T3D_EULER, T3D_ALLOW_CONVERSION);
  t3dInit(&bodyToRobotRed, T3D_EULER, T3D_ALLOW_CONVERSION);

  bodyToRobotRed.euler.euler[0] = 0.0;  /* theta */
  bodyToRobotRed.euler.euler[1] = pitch;
  bodyToRobotRed.euler.euler[2] = roll;
  bodyToRobotRed.euler.euler[3] = 0.0;  /* x */
  bodyToRobotRed.euler.euler[4] = 0.0;  /* y */
  bodyToRobotRed.euler.euler[5] = zBody;  /* z(Robot) ? */

  /* Coordinates of A,B,C,D in RobotRed */
  /* A */
  t3dInit(&pointToBody, T3D_EULER, T3D_ALLOW_CONVERSION);
  pointToBody.euler.euler[3] = length / 2.0;
  pointToBody.euler.euler[4] = width / 2.0;
  t3dCompIn(&pointToRobotRed, &pointToBody, &bodyToRobotRed);
  t3dConvertTo(T3D_EULER, &pointToRobotRed);
  xA = pointToRobotRed.euler.euler[3];
  yA = pointToRobotRed.euler.euler[4];

  /* D */
  t3dInit(&pointToBody, T3D_EULER, T3D_ALLOW_CONVERSION);
  pointToBody.euler.euler[3] = length / 2.0;
  pointToBody.euler.euler[4] = -width / 2.0;
  t3dCompIn(&pointToRobotRed, &pointToBody, &bodyToRobotRed);
  t3dConvertTo(T3D_EULER, &pointToRobotRed);
  xD = pointToRobotRed.euler.euler[3];
  yD = pointToRobotRed.euler.euler[4];

  /* B */
  t3dInit(&pointToBody, T3D_EULER, T3D_ALLOW_CONVERSION);
  pointToBody.euler.euler[3] = -length / 2.0;
  pointToBody.euler.euler[4] = width / 2.0;
  t3dCompIn(&pointToRobotRed, &pointToBody, &bodyToRobotRed);
  t3dConvertTo(T3D_EULER, &pointToRobotRed);
  xB = pointToRobotRed.euler.euler[3];
  yB = pointToRobotRed.euler.euler[4];

  /*-- bitmap size --*/
  /* along y */
  w = fabs(yD-yA);
  /* along x */
  h = fabs(xA-xB);

  /* create storage */
  body = malloc(sizeof(*body));
  if (body == NULL) { errno = ENOMEM; return NULL; }

  body->w = ceil(w/scale);
  body->h = ceil(h/scale);
  if (body->w < 1 || body->h < 1) { 
      errno = EINVAL; free(body);
      return NULL;
   }

  body->height = malloc(body->w * body->h * sizeof(*body->height));
   if (body->height == NULL) { errno = ENOMEM; free(body); return NULL; }

   body->scale = scale;
   
   /* center of body */
   t3dInit(&pointToBody, T3D_EULER, T3D_ALLOW_CONVERSION);
   pointToBody.euler.euler[3] = 0.0;
   pointToBody.euler.euler[4] = 0.0;
   t3dCompIn(&pointToRobotRed, &pointToBody, &bodyToRobotRed);
   t3dConvertTo(T3D_EULER, &pointToRobotRed);
   body->x = pointToRobotRed.euler.euler[3];
   body->y = pointToRobotRed.euler.euler[4];
   /* Verifier que c'est bien le bon (x,y) */

   /*---------------- fill height map now --------------------------*/
   body->nElem = 0;
   height = body->height;
   
   /* compute x and y scales on bdy plane */
   scale_x = width / ( (double) body->h );
   scale_y = length / ( (double) body->w );

   for(i=0; i<body->h; i++) {  /* along x */
     
     /* radial position between [-width/2;width/2] (meters) */
     r = width/2.0*(2.0*i/(body->h-1) - 1.0);  /* to be checked */

     for(j=0; j<body->w; j++, height++) {   /* along y */
       
       /* horizontal position between [-length/2;length/2] (meters) */
       l = length/2.0*(2.0*i/(body->w-1) - 1.0);  /* to be checked */
       
       /* altitude if undef by default */
       z = P3D_UNDEF_ELEM;

       /* compute height for this point */
       /*  (coord: (x,y,z)=(r,l,0) in RobotRed ) */
	   t3dInit(&pointToBody, T3D_EULER, T3D_ALLOW_CONVERSION);
       pointToBody.euler.euler[3] = r;
       pointToBody.euler.euler[4] = l;
       pointToBody.euler.euler[5] = 0.0;

       /* Compute current point position in RobotRed to get the z */
       t3dCompIn(&pointToRobotRed, &pointToBody, &bodyToRobotRed);
       t3dConvertTo(T3D_EULER, &pointToRobotRed);
       
       z = pointToRobotRed.euler.euler[5];

       /* fixed point height */
       if (z == P3D_UNDEF_ELEM)
	 *height = P3D_UNDEF_ELEM;
       else {
	 *height = z;
	 body->nElem++;
       }
     }
   }

   /* Put something to check the result */

   return body;
}

static void 
write_P3DElement(FILE *fp, const P3DElement* el)
{
	int i;
	fprintf(fp, "%d %d %f %f %f %d\n", el->w, el->h, el->x, el->y, 
			el->scale, el->nElem);

	for (i = 0; i < el->w * el->h; i++) {
		fprintf(fp, "%f ", el->height[i]);
	}	

	fprintf(fp,"\n");
}

static P3DElement* 
read_P3DElement(FILE *fp)
{
	int i;
	P3DElement* el;

	el = calloc(1, sizeof(*el));
	if (el == NULL) {
		fprintf(stderr, "%s : Can't allocate memory for P3DElement\n",__func__);
		return NULL;
	}

	fscanf(fp, "%d %d %f %f %f %d\n", &el->w, &el->h, &el->x, &el->y,
			&el->scale, &el->nElem);
	
	el->height = malloc(el->w * el->h * sizeof(*el->height));
	if (el->height == NULL) {
		fprintf(stderr, "%s : can't allocate memory for height\n", __func__);
		free(el);
		return NULL;
	}

	for (i = 0; i < el->w * el->h; i++) {
		fscanf(fp, "%f ", &el->height[i]);
	}

	// consumme eol
	fscanf(fp,"\n");

	return el;
}

void
p3dWriteP3DRobot(FILE *fp, const  P3DRobot *robot)
{
	int i, j;
	fprintf(fp, "%lf %d %lf %lf %lf %lf %lf %lf %d\n",
			robot->damping, robot->nIter,
		   	robot->rollMin, robot->rollStep,
			robot->bodyPitchMin, robot->bodyPitchStep,
			robot->bodyRollMin, robot->bodyRollStep,
			robot->nProfiles);

	for (i = 0; i < robot->nProfiles; i++) {
		write_P3DElement(fp, robot->lWheel[i]);
		write_P3DElement(fp, robot->rWheel[i]);
	}

#if defined(ATRV) || defined(ATRV_JUNIOR) || defined(TROOPER) || defined(RMP400)
	fprintf(fp, "%d %d %d\n", robot->nBodyProfiles, robot->nBodyRoll, robot->nBodyPitch);
	for (i = 0; i < robot->nBodyRoll; i++) {
		for (j=0; j < robot->nBodyPitch; j++) {
			write_P3DElement(fp, robot->body[i][j]);
		}
	}
#endif

	fprintf(fp,"%lf %lf %lf %lf\n", robot->vMin, robot->vMax,
		   	robot->accel, robot->decel);
}

P3DRobot *
p3dReadP3DRobot(FILE *fp)
{
	int i, j;
	P3DRobot *robot = calloc(1, sizeof(*robot));
	if (robot == NULL) {
		fprintf(stderr, "%s : can't allocate memory for robot\n", __func__);
		return NULL;
	}

	fscanf(fp, "%lf %d %lf %lf %lf %lf %lf %lf %d\n",
			&robot->damping, &robot->nIter,
		   	&robot->rollMin, &robot->rollStep,
			&robot->bodyPitchMin, &robot->bodyPitchStep,
			&robot->bodyRollMin, &robot->bodyRollStep,
			&robot->nProfiles);


	robot->lWheel = calloc(robot->nProfiles, sizeof(*robot->lWheel));
	if (robot->lWheel == NULL) {
		fprintf(stderr, "%s : can't allocate memory for lWheel\n", __func__);
		p3dRobotDestroy(robot);
		return NULL;
	}

	robot->rWheel = calloc(robot->nProfiles, sizeof(*robot->rWheel));
	if (robot->rWheel == NULL) {
		fprintf(stderr, "%s : can't allocate memory for rWheel\n", __func__);
		p3dRobotDestroy(robot);
		return NULL;
	}

	for (i = 0; i < robot->nProfiles; i++) {
		robot->lWheel[i] = read_P3DElement(fp);
		robot->rWheel[i] = read_P3DElement(fp);
		if (robot->lWheel[i] == NULL || robot->rWheel[i] == NULL) {
			p3dRobotDestroy(robot);
			return NULL;
		}
	}

#if defined(ATRV) || defined(ATRV_JUNIOR) || defined(TROOPER) || defined(RMP400)
	fscanf(fp, "%d %d %d\n", &robot->nBodyProfiles, &robot->nBodyRoll, &robot->nBodyPitch);

	robot->body = calloc(robot->nBodyRoll, sizeof(*robot->body));
	if (robot->body == NULL) { 
		fprintf(stderr, "%s : can't allocate memory for body\n", __func__);
		p3dRobotDestroy(robot);
		return NULL;
	}

	for(i=0; i< robot->nBodyRoll; i++)
	{
		robot->body[i] = calloc(robot->nBodyPitch, sizeof(**robot->body));
		if (robot->body[i] == NULL) { 
			fprintf(stderr, "%s : can't allocate memory for body\n", __func__);
			p3dRobotDestroy(robot);
			return NULL;
		}
	}

	for (i = 0; i < robot->nBodyRoll; i++) {
		for (j=0; j < robot->nBodyPitch; j++) {
			robot->body[i][j] = read_P3DElement(fp);
			if (robot->body[i][j] == NULL) {
				p3dRobotDestroy(robot);
				return NULL;
			}
		}
	}
#endif

	fscanf(fp,"%lf %lf %lf %lf\n", &robot->vMin, &robot->vMax,
		   	&robot->accel, &robot->decel);

	return robot;
}
