/**
 ** David Bonnafous - Anthony Mallet - Thu Jul  5 2001
 ** Copyright (c) 2001 LAAS/CNRS 
 **
 ** $Log: search.c,v $
 ** Revision 1.4  2007/11/27 10:41:21  adegroot
 ** Clean a bit the code so it can compile with -W -Wall -Wextra -Wwrite-strings -Werror
 **
 ** -pedantic is not supported because we use gnu extension #ident
 **
 ** Revision 1.3  2007/11/22 09:58:37  adegroot
 ** In solution arc, store the real full path of the arc. Previously we miss the
 ** last element of the path.
 **
 ** Hopefully, the cost, bestCurv, and ds were correctly computed.
 **
 ** Revision 1.2  2007/09/10 10:55:23  sjoyeux
 ** return the found path instead than the first arc of the path
 **
 **   Instead of returning only the first arc, the search functions now assume
 **   that bestArc is an array if P3DArc structures of at least depth elements.
 **   This array is filled with the arcs composing the path, starting by the
 **   arc nearest to the goal.
 **
 **   The new output argument solutionDepth contains the number of arcs actually
 **   found in the solution path.
 **
 ** Revision 1.1.1.1  2006/10/17 18:09:42  sjoyeux
 ** Packaged the P3D module from Lama's calife.
 ** See CHANGES for a list of modifications
 **
 **
 ** Revision 1.6.2.5  2005/07/13 15:50:12  tpeynot
 ** add some printf for error cases that did not have
 **
 ** Revision 1.6.2.4  2005/04/26 12:26:51  tpeynot
 ** conf.x.y.z were not copied in bestArc
 **
 ** Revision 1.6.2.3  2005/04/13 20:27:39  tpeynot
 ** Modifications of protos and structures for body collision test (e.g. ground clearance info)
 **
 ** Revision 1.6.2.2  2005/03/24 23:28:38  tpeynot
 ** Copy bestArc Info where it is available
 **
 ** Revision 1.6.2.1  2005/03/18 23:42:39  tpeynot
 ** p3d uses for now an internal map: a struct P3D_MAP
 ** This struct will contain info a dem from the dem module
 **
 ** Revision 1.6  2005/03/17 13:46:54  tpeynot
 ** Add param "bestArc" ro p3dPlan fct to get info about the arc selectes by p3d planner
 ** Add functions adapted to Dala's structure instead of Lama. These functions are used if there's a flag -DATRV in Makefile. Otherwise, the lib will keep working with Lama adapted functions
 **
 ** Revision 1.5  2003/09/12 15:21:02  mallet
 ** correctly handle arcs that cross goalTol limit
 **
 ** Revision 1.4  2002/10/09 15:45:05  mallet
 ** p3dPlan retourne le cout estime, normalise entre 1 et +infini
 **
 ** Revision 1.3  2001/09/13 10:50:29  mallet
 ** Heuristique qui permet de ne pas considerer la zone aveugle a l'init
 **
 ** Revision 1.2  2001/07/14 11:39:09  mallet
 ** premiere version ok
 **
 ** Revision 1.1  2001/07/12 17:40:03  mallet
 ** sync
 **
 ** Revision 1.1  2001/07/11 18:05:40  mallet
 ** re-ecriture de p3d
 **
 **/
#ident "$Id: search.c,v 1.4 2007/11/27 10:41:21 adegroot Exp $"

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include <assert.h>

#include "libp3d.h"
#include "p3dInt.h"
#include "avl.h"
#include "p3dFullImpl.h"

static int	p3dOpenNode(P3DPlanner *planner, P3D_MAP *map,
			double x, double y, double theta, P3DArc *arcs,
			int nArcs, int node, AVLNode **open);
static int	p3dOrderCost(void *a, void *b);

/*
 * p3dAStar -------------------------------------------------------------
 *
 * Find best arc
 */

int
p3dAStar(P3DPlanner *planner, P3D_MAP *map,
	 double x, double y, double theta,
	 double gx, double gy, double goalTol,
	 void (*callback)(P3DNode *node, void *data), void *data,
	 P3DPlanSolution *sol)
{
   P3DArc *arc;	
   P3DNode *best;
   AVLNode *open = NULL;
   double stop_distance;

   p3dResetPlanSolution(sol);

   /* It may be better to compute stop_distance / arcs->max_depth */
   stop_distance = MAX(p3dGetStopDistance(sol->v, planner), planner->dMin);

   /* 
	* Recompute the node position with the new stop_distance 
	* it may be better to only recompute if stop_distance_previous != stop_distance
	* */
   p3dArcComputeNodePosition(planner->arcs, planner->nArcs, 0.0, 0.0, 0.0, stop_distance);

   /* reset costs and compute distance to goal for each config */
   p3dArcsReset(planner->arcs, planner->nArcs, x, y, theta, gx, gy, goalTol); 
   
   /* open first nodes */
   if (p3dOpenNode(planner, map, x, y, theta, 
		   planner->arcs, planner->nArcs, 0, &open)) return -1;

   /* search */
   best = NULL;
   while (open != NULL) {

      /* get best node and invoke callback */
      best = AVLFindDeleteFirstNode(&open);
      if (callback != NULL) callback(best, data);

      /* stop on arc as soon as we've reached a position which is a local min */
      if (best->isLocalMin) break;

      /* open children */
      arc = best->parent;
      if (best->number >= arc->n-1) {
	 if (arc->nArcs == 0) break;
	 if (p3dOpenNode(planner, map, x, y, theta, 
			 arc->arcs, arc->nArcs, 0, &open)) break;
      } else {
	 if (p3dOpenNode(planner, map, x, y, theta, 
			 arc, 1, best->number+1, &open)) break;
      }
   }

   if (best == NULL) {
     fprintf(stderr, "p3dAStar: best == NULL\n");
     return -1;
   }

   p3dSetPlanSolution(sol, planner, best->parent);

   return 0;
}


/*
 * p3dOpenNode ----------------------------------------------------------
 *
 * Valuate a node and insert it in "open"
 */

static int
p3dOpenNode(P3DPlanner *planner, P3D_MAP *map,
	    double x, double y, double theta,
	    P3DArc *arcs, int nArcs, int node,
	    AVLNode **open)
{
   double cost = cos(theta), sint = sin(theta);
   double xn, yn, thetan;
   double q;
   P3DRobotConfig start, *prev;
   P3DArc *a;
   int i, j;

   /* place robot at origin for the very first node only */
   if (node == 0 && arcs->parent == NULL) {
      q = PLACE_ROBOT(planner->robot, planner->geom, x, y, theta, map, &start);
      if (q < 0.0) {
	fprintf(stderr, "p3dOpenNode: Failed to place robot at origin\n");
	return -1;
      }
   }

   for(i=0, a=arcs; i<nArcs; i++, a++) {

      if (node == 0) {
	 /* set the cost of this arc to the one of its parent */
	 a->cost = (a->parent != NULL) ? a->parent->cost : 0.0;
      }
      a->current = node;
 
      /* valuate node */
      if (a->nodes[node].cost == P3D_UNDEF_COST) {
	 xn = a->nodes[node].x*cost - a->nodes[node].y*sint + x;
	 yn = a->nodes[node].x*sint + a->nodes[node].y*cost + y;
	 thetan = a->nodes[node].theta + theta;

	 /* compute current configuration */
	 q = PLACE_ROBOT(planner->robot, planner->geom,
			   xn, yn, thetan, map, &a->nodes[node].conf);
	 if (q < 0.0) continue;

	 /* Check for unknow node */
	 check_unknow(planner->geom, a, node);

	 /* and previous one */
	 prev = a->nodes[node].prev;
	 if (prev == NULL) prev = &start;

	 a->nodes[node].cost = a->ds *
	   EVAL_CONFIG(planner->geom, &a->nodes[node].conf, prev);

	 if (a->nodes[node].cost < 0.0) continue;

	 /* remove unknown at the beginning */
	 if (a->parent == NULL && 
	     a->nodes[node].conf.pUnk > planner->geom->unkMax) {
	    for(j=0;j<node;j++) {
	       if (a->nodes[node].conf.pUnk <= planner->geom->unkMax)
		  break;
	    }
	    if (j >= node) {
	       a->nodes[node].cost = a->ds;
	    }
	 }
      }

      a->cost += a->nodes[node].cost;

      /* insert node in open */
      AVLInsertNode(open, &a->nodes[node], p3dOrderCost);
   }

   return 0;
}


/*
 * p3dOrderCost ---------------------------------------------------------
 *
 * Sort function for avls
 */

static int
p3dOrderCost(void *a, void *b)
{
   P3DNode *m = a, *n = b;

   if (m->parent->cost + m->remain <= n->parent->cost + n->remain)
      return L_BAL;
   else
      return R_BAL;
}

/*
 * p3dResetPlanSolution -------------------------------------------------
 *
 * Set the plan Solution as currently indetermined
 */
void 
p3dResetPlanSolution(P3DPlanSolution *sol)
{
	sol->is_valid = 0;
	sol->depth = 0;
}

/*
 * p3dSetPlanSolution ----------------------------------------------------
 *
 * For a best arc, set the solution
 */
void
p3dSetPlanSolution(P3DPlanSolution *sol, P3DPlanner * planner, P3DArc *best)
{
	int k;
	P3DArc *arc = NULL;
	P3DArc *currentBest = NULL;
	double speed;

	/* find root arc and fill the solution return value */
	for(arc = best; arc != NULL; arc = arc->parent, ++(sol->depth))
	{
		currentBest = sol->bestArc + sol->depth;
		currentBest->r = arc->r;
		currentBest->ds = arc->ds;
		currentBest->depth = arc->depth;
		currentBest->n = arc->n;
		currentBest->current = arc->current;
		currentBest->first_unknow = arc->first_unknow;
		currentBest->cost = arc->cost;
		arc->is_elected = 1;

		/* Copy information in each node */
		for(k=0; k < currentBest->n; k++)
		{
			currentBest->nodes[k].x = arc->nodes[k].x;
			currentBest->nodes[k].y = arc->nodes[k].y;
			currentBest->nodes[k].theta = arc->nodes[k].theta;

			/* Configurations */
			currentBest->nodes[k].conf.q = arc->nodes[k].conf.q;
			currentBest->nodes[k].conf.x = arc->nodes[k].conf.x;
			currentBest->nodes[k].conf.y = arc->nodes[k].conf.y;
			currentBest->nodes[k].conf.z = arc->nodes[k].conf.z;
			currentBest->nodes[k].conf.theta = arc->nodes[k].conf.theta;
			currentBest->nodes[k].conf.dzBodyGround = arc->nodes[k].conf.dzBodyGround;
			currentBest->nodes[k].conf.roll = arc->nodes[k].conf.roll;
			currentBest->nodes[k].conf.pitch = arc->nodes[k].conf.pitch;
			currentBest->nodes[k].conf.alpha1 = arc->nodes[k].conf.alpha1;
			currentBest->nodes[k].conf.alpha2 = arc->nodes[k].conf.alpha2;
			currentBest->nodes[k].conf.beta3 = arc->nodes[k].conf.beta3;
			currentBest->nodes[k].conf.pUnk = arc->nodes[k].conf.pUnk;

			currentBest->nodes[k].cost = arc->nodes[k].cost;
			currentBest->nodes[k].remain= arc->nodes[k].remain;
		}
	}

	/* these values are for the first arc only */
	sol->curv = currentBest->r;
	sol->d = currentBest->ds * (currentBest->current+1);

	if (sol->d > 0.)
		sol->cost = currentBest->cost / sol->d; /* (renormalize) */
	else
		sol->cost = 1.;

	/* There is a simple relation between r and alpha 
	 * r  = alpha / L where L is the length between front axle and fix axle
	 * (back axle for dala, middle axle for Trooper
	 * For Lama, we can't compute it but does it really make sense ? Most of
	 * the code can't work for it because we use an Ackerman model and lama
	 * doesn't fit in this model.
	 * This must be revisited when we use a new robot like lama (if we do it
	 * some day)
	 */
	sol->alpha = sol->curv * p3dGetDistanceFixAndMobileAxle(planner);

	/* To compute v, use a simple heuristic for the  moment
	 * Use a linear function based on cost
	 *
	 * sol->v = planner->robot->vMin + 
	 *	( planner->robot->vMax - planner->robot->vMin ) / sol->cost;
	 * XXX As the cost really increased, the previous heuristic is incorrect,
	 * so, for moment, just use vMax. It is completly incorrect and must be
	 * revisited, on the base of the difficulty of the path
	 */
	sol->v = planner->robot->vMax;

	speed = max_speed(currentBest, planner);
	if (sol->v > speed) {
		sol->v = speed;
	};

	/* Compute sol->w (and sol->v) to 
	 * 1 : realize the path associated to curv
	 * 2 : respect v and w values domain
	 */
	if (fabs(sol->curv) < 1e-5) {
		sol->w = 0;
	} else {
		int sign;
		if (sol->curv > 0) 
			sign = +1;
		else
			sign = -1;
		double r = fabs(1 / sol->curv);
		if (sol->v > r * planner->robot->wMax) {
			sol->w = sign * planner->robot->wMax;
			sol->v = r * planner->robot->wMax;
		} else {
			sol->w = sign * sol->v / r;
		}
	}

	sol->is_valid = 1;
}


int
p3dFull(P3DPlanner *planner, P3D_MAP *map,
	 double x, double y, double theta,
	 double gx, double gy, double goalTol,
	 void (*callback)(P3DNode *node, void *data), void *data,
	 P3DPlanSolution *sol)
{
   const P3DArc* best; 
   double stop_distance;
   (void) callback;
   (void) data;

   p3dResetPlanSolution(sol);

   /* It may be better to compute stop_distance / arcs->max_depth */
   stop_distance = MAX(p3dGetStopDistance(sol->v, planner), planner->dMin);

   /* 
	* Recompute the node position with the new stop_distance 
	* it may be better to only recompute if stop_distance_previous != stop_distance
	* */
   p3dArcComputeNodePosition(planner->arcs, planner->nArcs, 0.0, 0.0, 0.0, stop_distance);

   /* reset costs and compute distance to goal for each config */
   p3dArcsReset(planner->arcs, planner->nArcs, x, y, theta, gx, gy, goalTol); 

   /* Compute the cost of each arcs */
   p3dArcComputeCost(planner, map, planner->arcs, planner->nArcs, x, y, theta, goalTol);

   /* Pondate the cost of each arcs */
   p3dArcPonderateCost(planner->arcs, planner->nArcs);

   /* Elected the best arc */
   best = p3dElectBestArc(planner->arcs, planner->nArcs);

   if (best == NULL) {
     fprintf(stderr, "p3dFull: best == NULL\n");
     return -1;
   }

   /* 
	* XXX so crap, best must be modified to show that it is elected, maybe we
	* need to remove this 'not so useful feature'
	*/
   p3dSetPlanSolution(sol, planner, (P3DArc*) best);

   return 0;
}

/*
 * P3dArcComputeCostLimited  ----------------------------------------------------------
 *
 * Compute the cost of any nodes / arcs. If we find something impossible on an
 * node, stop the computation here and affect the impossible cost to the arc
 */
static int 
_p3dArcComputeCostLimited(P3DPlanner *planner, P3D_MAP *map, P3DArc *arcs, int n, 
				  double x, double y, double theta, double goalTol, double min, double max,
				  P3DRobotConfig *start)
{
	double cost = cos(theta), sint = sin(theta);
	double xn, yn, thetan;
	double q;
	P3DRobotConfig *prev;
	P3DArc *a;
	int i, j;
	bool possible;

	for (i = 0, a = arcs; i<n; i++, a++) {
		if (a->r < min || a->r > max) {
			fprintf(stderr, "Don't compute this arc, a->r = %lf is out of limits\n", a->r);
			continue;
		}
		j = 0;
		possible = true;

		/* Init a->cost */
		a->cost = (a->parent != NULL) ? a->parent->cost : 0.0;

		/* Don't compute the son arc if we can't survive to the parent */
		if (a->cost > P3D_IMPOSSIBLE_COST) 
			return 0;

		/* Compute cost of each node */
		while (j < a->n && possible) {
			a->current = j;

			xn = a->nodes[j].x*cost - a->nodes[j].y*sint + x;
			yn = a->nodes[j].x*sint + a->nodes[j].y*cost + y;
			thetan = a->nodes[j].theta + theta;

			/* compute current configuration */
			q = PLACE_ROBOT(planner->robot, planner->geom,
					xn, yn, thetan, map, &a->nodes[j].conf);
			if (q < 0.0) {
				a->nodes[j].cost = P3D_IMPOSSIBLE_COST;
				possible = false;
			} else {
				/* check for unknow node */
				check_unknow(planner->geom, a, j);

				/* and previous one */
				prev = a->nodes[j].prev;
				if (prev == NULL) prev = start;

				/* compute its cost */
				a->nodes[j].cost = a->ds *
					EVAL_CONFIG(planner->geom, &a->nodes[j].conf, prev);

				if (a->nodes[j].cost < 0.0) {
					a->nodes[j].cost = P3D_IMPOSSIBLE_COST;
					possible = false;
				}
			}
			a->cost += a->nodes[j].cost;

			/* If we are near the goal, don't compute more */
			if (a->nodes[j].remain < goalTol) {
				possible = false;
			}
			j++;
		}
		
		/* Do it recursivly if necessary (but it will cost a lot */
		if (a->nArcs != 0) {
			_p3dArcComputeCostLimited(planner, map,
					a->arcs, a->nArcs,
					x, y, theta,
					goalTol,
					min,
					max,
					&a->nodes[a->n - 1].conf);
		}
	}

	return 0;
}

int 
p3dArcComputeCostLimited(P3DPlanner *planner, P3D_MAP *map, P3DArc *arcs, int n, 
				  double x, double y, double theta, double goalTol, double min, double max)
{
	P3DRobotConfig start;
	double q;

	/* Compute start configuration */
	q = PLACE_ROBOT(planner->robot, planner->geom, x, y, theta, map, &start);
	if (q < 0.0) {
		fprintf(stderr, "p3dArcComputeCost : Failed to place robot at origin\n");
		return -1;
	}

	/* Be sure that all flags are reset before modifying them in _p3dArcComputeCostLimited */
	p3dArcsResetFlags(arcs, n);
	
	return _p3dArcComputeCostLimited(planner, map, arcs, n, x, y, theta, goalTol, min, max, &start);
}

/*
 * Ponderate the cost of each arc with the ones around it. For moment, consider
 * only its two neighbour but we can use in the future more complex scheme.
 */
void
p3dArcPonderateCost(P3DArc * arcs, int n)
{
	P3DArc *a;
	int i;

	/* Get the balanced cost of the parent */
	for (i = 0, a = arcs; i < n; i++, a++) {
		if (a->parent != NULL) {
			a->cost = a->parent->cost_balanced;
		}
	}

	/* 
	 * Special case for first arc 
	 * We consider the unknow arc is easy
	 * The right thing is probably to compute two extra arcs and don't include
	 * it in the computing.
	 */
	arcs->cost_balanced = fPonderation(arcs->cost, EASY * arcs->n * arcs->ds, (arcs+1)->cost); 
		
	for (i = 1, a =arcs+1; i < (n-1); i++, a++) 
		a->cost_balanced = fPonderation(a->cost, (a-1)->cost, (a+1)->cost);


	/* 
	 * Special case for last case
	 * See the previous comment
	 */
	a = & arcs[n-1];
	a->cost_balanced = fPonderation(a->cost, (a-1)->cost, EASY * a->n * a->ds);

	/*
	 * Do it recursively if necessary 
	 */
	for (i = 0, a = arcs; i < n; i++, a++) {
		if (a->nArcs != 0) {
			p3dArcPonderateCost(a->arcs, a->nArcs);
		}
	}
}

static void _p3dElectBestArc(const P3DArc*, int, double*, const P3DArc **);

const P3DArc*
p3dElectBestArc(const P3DArc *arcs, int n)
{
	const P3DArc *res = NULL;
	double best_score = P3D_IMPOSSIBLE_COST;

	_p3dElectBestArc(arcs, n, &best_score, &res);

	/* Be sure that res is acceptable 
	 * If res is an impossible path, there is no available path
	 * */
	if (res != NULL && (res->cost >= P3D_IMPOSSIBLE_COST))
		res = NULL;
	return res;
}

static void
_p3dElectBestArc(const P3DArc *arcs, int n, double * best_score, const P3DArc **res)
{
	int i;
	const P3DArc *a;

	for (i = 0, a = arcs; i < n; i++, a++) {
		double score = p3dFullComputeScore(a);
		if (score < *best_score) {
			*best_score = score; 
			*res = a;
		}

		if (a->nArcs != 0) {
			_p3dElectBestArc(a->arcs, a->nArcs, best_score, res);
		}
	}
}

double
fPonderation(double current_cost, double previous_cost, double next_cost)
{
	return 0.5 * (current_cost + 0.5 * (previous_cost + next_cost));
}
